"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.powerpath as powerpath
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {powerpath: {}}


def test_has_powerpath():
    """
    Test for powerpath
    """
    with patch("os.path.exists") as mock_exists:
        mock_exists.return_value = True
        assert powerpath.has_powerpath()

        mock_exists.return_value = False
        assert not powerpath.has_powerpath()


def test_list_licenses():
    """
    Test to returns a list of applied powerpath license keys
    """
    with patch.dict(powerpath.__salt__, {"cmd.run": MagicMock(return_value="A\nB")}):
        assert powerpath.list_licenses() == []


def test_add_license():
    """
    Test to add a license
    """
    with patch.object(powerpath, "has_powerpath", return_value=False):
        assert powerpath.add_license("key") == {
            "output": "PowerPath is not installed",
            "result": False,
            "retcode": -1,
        }

    mock = MagicMock(return_value={"retcode": 1, "stderr": "stderr"})
    with patch.object(powerpath, "has_powerpath", return_value=True):
        with patch.dict(powerpath.__salt__, {"cmd.run_all": mock}):
            assert powerpath.add_license("key") == {
                "output": "stderr",
                "result": False,
                "retcode": 1,
            }


def test_remove_license():
    """
    Test to remove a license
    """
    with patch.object(powerpath, "has_powerpath", return_value=False):
        assert powerpath.remove_license("key") == {
            "output": "PowerPath is not installed",
            "result": False,
            "retcode": -1,
        }

    mock = MagicMock(return_value={"retcode": 1, "stderr": "stderr"})
    with patch.object(powerpath, "has_powerpath", return_value=True):
        with patch.dict(powerpath.__salt__, {"cmd.run_all": mock}):
            assert powerpath.remove_license("key") == {
                "output": "stderr",
                "result": False,
                "retcode": 1,
            }



"""
Management of nftables
======================

This is an nftables-specific module designed to manage Linux firewalls. It is
expected that this state module, and other system-specific firewall states, may
at some point be deprecated in favor of a more generic `firewall` state.

.. code-block:: yaml

    httpd:
      nftables.append:
        - table: filter
        - chain: input
        - jump: accept
        - match: state
        - connstate: new
        - dport: 80
        - proto: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      nftables.append:
        - table: filter
        - family: ipv6
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - proto: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      nftables.insert:
        - position: 1
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - proto: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      nftables.insert:
        - position: 1
        - table: filter
        - family: ipv6
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - proto: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      nftables.delete:
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - proto: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      nftables.delete:
        - position: 1
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - proto: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      nftables.delete:
        - table: filter
        - family: ipv6
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - proto: tcp
        - sport: 1025:65535
        - save: True

    output:
      nftables.chain_present:
        - family: ip
        - table: filter

    output:
      nftables.chain_absent:
        - family: ip
        - table: filter

"""

import logging

from salt.state import STATE_INTERNAL_KEYWORDS as _STATE_INTERNAL_KEYWORDS

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the locale module is available in __salt__
    """
    if "nftables.version" in __salt__:
        return "nftables"
    return (False, "nftables module could not be loaded")


def chain_present(
    name, table="filter", table_type=None, hook=None, priority=None, family="ipv4"
):
    """
    .. versionadded:: 2014.7.0

    .. versionchanged:: 3002

    Verify a chain exists in a table.

    name
        A user-defined chain name.

    table
        The table to own the chain.

    family
        Networking family, either ipv4 or ipv6
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    chain_check = __salt__["nftables.check_chain"](table, name, family=family)
    if chain_check["result"] is True:
        ret["result"] = True
        ret["comment"] = "nftables {} chain is already exist in {} table for {}".format(
            name, table, family
        )
        return ret

    if __opts__["test"]:
        ret["comment"] = (
            "nftables chain {} would be created in table {} for family {}".format(
                name, table, family
            )
        )
        return ret

    res = __salt__["nftables.new_chain"](
        table, name, table_type=table_type, hook=hook, priority=priority, family=family
    )

    if res["result"] is True:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "nftables {} chain in {} table create success for {}".format(
            name, table, family
        )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = "Failed to create {} chain in {} table: {} for {}".format(
            name, table, res["comment"].strip(), family
        )
        return ret


def chain_absent(name, table="filter", family="ipv4"):
    """
    .. versionadded:: 2014.7.0

    Verify the chain is absent.

    family
        Networking family, either ipv4 or ipv6
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    chain_check = __salt__["nftables.check_chain"](table, name, family)
    if not chain_check:
        ret["result"] = True
        ret["comment"] = (
            "nftables {} chain is already absent in {} table for {}".format(
                name, table, family
            )
        )
        return ret

    flush_chain = __salt__["nftables.flush"](table, name, family)
    if flush_chain:
        command = __salt__["nftables.delete_chain"](table, name, family)
        if command is True:
            ret["changes"] = {"locale": name}
            ret["result"] = True
            ret["comment"] = (
                "nftables {} chain in {} table delete success for {}".format(
                    name, table, family
                )
            )
        else:
            ret["result"] = False
            ret["comment"] = "Failed to delete {} chain in {} table: {} for {}".format(
                name, table, command.strip(), family
            )
    else:
        ret["result"] = False
        ret["comment"] = "Failed to flush {} chain in {} table: {} for {}".format(
            name, table, flush_chain.strip(), family
        )
    return ret


def append(name, family="ipv4", **kwargs):
    """
    .. versionadded:: 0.17.0

    Append a rule to a chain

    name
        A user-defined name to call this rule by in another part of a state or
        formula. This should not be an actual rule.

    family
        Network family, ipv4 or ipv6.

    All other arguments are passed in with the same name as the long option
    that would normally be used for nftables, with one exception: `--state` is
    specified as `connstate` instead of `state` (not to be confused with
    `ctstate`).
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]
    res = __salt__["nftables.build_rule"](family=family, **kwargs)
    if not res["result"]:
        return res
    rule = res["rule"]

    res = __salt__["nftables.build_rule"](
        full=True, family=family, command="add", **kwargs
    )
    if not res["result"]:
        return res
    command = res["rule"]

    res = __salt__["nftables.check"](kwargs["table"], kwargs["chain"], rule, family)
    if res["result"]:
        ret["result"] = True
        ret["comment"] = "nftables rule for {} already set ({}) for {}".format(
            name, command.strip(), family
        )
        return ret
    if "test" in __opts__ and __opts__["test"]:
        ret["comment"] = "nftables rule for {} needs to be set ({}) for {}".format(
            name, command.strip(), family
        )
        return ret
    res = __salt__["nftables.append"](kwargs["table"], kwargs["chain"], rule, family)
    if res["result"]:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "Set nftables rule for {} to: {} for {}".format(
            name, command.strip(), family
        )
        if "save" in kwargs:
            if kwargs["save"]:
                __salt__["nftables.save"](filename=None, family=family)
                ret["comment"] = (
                    "Set and Saved nftables rule for {} to: {} for {}".format(
                        name, command.strip(), family
                    )
                )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = (
            "Failed to set nftables rule for {}.\nAttempted rule was {} for {}.\n{}".format(
                name, command.strip(), family, res["comment"]
            )
        )
        return ret


def insert(name, family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.7.0

    Insert a rule into a chain

    name
        A user-defined name to call this rule by in another part of a state or
        formula. This should not be an actual rule.

    family
        Networking family, either ipv4 or ipv6

    All other arguments are passed in with the same name as the long option
    that would normally be used for nftables, with one exception: `--state` is
    specified as `connstate` instead of `state` (not to be confused with
    `ctstate`).
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]
    res = __salt__["nftables.build_rule"](family=family, **kwargs)
    if not res["result"]:
        return res
    rule = res["rule"]

    res = __salt__["nftables.build_rule"](
        full=True, family=family, command="insert", **kwargs
    )
    if not res["result"]:
        return res
    command = res["rule"]

    res = __salt__["nftables.check"](kwargs["table"], kwargs["chain"], rule, family)
    if res["result"]:
        ret["result"] = True
        ret["comment"] = "nftables rule for {} already set for {} ({})".format(
            name, family, command.strip()
        )
        return ret
    if "test" in __opts__ and __opts__["test"]:
        ret["comment"] = "nftables rule for {} needs to be set for {} ({})".format(
            name, family, command.strip()
        )
        return ret
    res = __salt__["nftables.insert"](
        kwargs["table"], kwargs["chain"], kwargs["position"], rule, family
    )
    if res["result"]:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "Set nftables rule for {} to: {} for {}".format(
            name, command.strip(), family
        )
        if "save" in kwargs:
            if kwargs["save"]:
                __salt__["nftables.save"](filename=None, family=family)
                ret["comment"] = (
                    "Set and Saved nftables rule for {} to: {} for {}".format(
                        name, command.strip(), family
                    )
                )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = (
            "Failed to set nftables rule for {}.\nAttempted rule was {}".format(
                name, command.strip()
            )
        )
        return ret


def delete(name, family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.7.0

    Delete a rule to a chain

    name
        A user-defined name to call this rule by in another part of a state or
        formula. This should not be an actual rule.

    family
        Networking family, either ipv4 or ipv6

    All other arguments are passed in with the same name as the long option
    that would normally be used for nftables, with one exception: `--state` is
    specified as `connstate` instead of `state` (not to be confused with
    `ctstate`).
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]
    res = __salt__["nftables.build_rule"](family=family, **kwargs)
    if not res["result"]:
        return res
    rule = res["rule"]

    res = __salt__["nftables.build_rule"](
        full=True, family=family, command="D", **kwargs
    )
    if not res["result"]:
        return res
    command = res["rule"]

    res = __salt__["nftables.check"](kwargs["table"], kwargs["chain"], rule, family)

    if not res["result"]:
        ret["result"] = True
        ret["comment"] = "nftables rule for {} already absent for {} ({})".format(
            name, family, command.strip()
        )
        return ret
    if "test" in __opts__ and __opts__["test"]:
        ret["comment"] = "nftables rule for {} needs to be deleted for {} ({})".format(
            name, family, command.strip()
        )
        return ret

    if "position" in kwargs:
        res = __salt__["nftables.delete"](
            kwargs["table"], kwargs["chain"], family=family, position=kwargs["position"]
        )
    else:
        res = __salt__["nftables.delete"](
            kwargs["table"], kwargs["chain"], family=family, rule=rule
        )

    if res["result"]:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = f"Delete nftables rule for {name} {command.strip()}"
        if "save" in kwargs:
            if kwargs["save"]:
                __salt__["nftables.save"](filename=None, family=family)
                ret["comment"] = (
                    "Deleted and Saved nftables rule for {} for {}{}".format(
                        name, command.strip(), family
                    )
                )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = (
            "Failed to delete nftables rule for {}.\nAttempted rule was {}".format(
                name, command.strip()
            )
        )
        return ret


def flush(name, family="ipv4", ignore_absence=False, **kwargs):
    """
    .. versionadded:: 2014.7.0

    .. versionchanged:: 3002

    Flush current nftables state

    family
        Networking family, either ipv4 or ipv6

    ignore_absence
        If set to True, attempts to flush a non-existent table will not
        result in a failed state.

        .. versionadded:: 3002

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if __opts__["test"]:
        ret["comment"] = "nftables flush not performed in test mode."
        return ret

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]

    if "table" not in kwargs:
        kwargs["table"] = "filter"

    check_table = __salt__["nftables.check_table"](kwargs["table"], family=family)
    if not ignore_absence and not check_table["result"]:
        ret["result"] = False
        ret["comment"] = (
            "Failed to flush table {} in family {}, table does not exist.".format(
                kwargs["table"], family
            )
        )
        return ret

    if "chain" not in kwargs:
        kwargs["chain"] = ""
    else:
        check_chain = __salt__["nftables.check_chain"](
            kwargs["table"], kwargs["chain"], family=family
        )
        if not ignore_absence and not check_chain["result"]:
            ret["result"] = False
            ret["comment"] = (
                "Failed to flush chain {} in table {} in family {}, chain does not"
                " exist.".format(kwargs["chain"], kwargs["table"], family)
            )
            return ret

    res = __salt__["nftables.flush"](kwargs["table"], kwargs["chain"], family)
    if res["result"] or (
        ignore_absence and (not check_table["result"] or not check_chain["result"])
    ):
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "Flush nftables rules in {} table {} chain {} family".format(
            kwargs["table"], kwargs["chain"], family
        )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = "Failed to flush nftables rules"
        return ret


def set_policy(name, table="filter", family="ipv4", **kwargs):
    """
    .. versionadded:: 3002

    Sets the default policy for nftables chains

    table
        The table that owns the chain that should be modified

    family
        Networking family, either ipv4 or ipv6

    policy
        The requested table policy (accept or drop)

    save
        Boolean to save the in-memory nftables settings to a file.

    save_filename
        The filename to save the nftables settings (default: /etc/nftables
        or /etc/nftables/salt-all-in-one.nft if the former is a directory)

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]

    policy = __salt__["nftables.get_policy"](table, kwargs["chain"], family)

    if (policy or "").lower() == kwargs["policy"].lower():
        ret["result"] = True
        ret["comment"] = (
            "nftables default policy for chain {} on table {} for {} already set to {}".format(
                kwargs["chain"], table, family, kwargs["policy"]
            )
        )
        return ret

    if __opts__["test"]:
        ret["comment"] = (
            "nftables default policy for chain {} on table {} for {} needs to be set"
            " to {}".format(kwargs["chain"], table, family, kwargs["policy"])
        )
        return ret

    if __salt__["nftables.set_policy"](
        table, kwargs["chain"], kwargs["policy"].lower(), family
    ):
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "Set default policy for {} to {} family {}".format(
            kwargs["chain"], kwargs["policy"], family
        )

        if "save" in kwargs:
            if kwargs["save"]:
                __salt__["nftables.save"](
                    filename=kwargs.get("save_filename"), family=family
                )
                ret["comment"] = (
                    "Set and saved default policy for {} to {} family {}".format(
                        kwargs["chain"], kwargs["policy"], family
                    )
                )
    else:
        ret["result"] = False
        ret["comment"] = "Failed to set nftables default policy"

    return ret


def table_present(name, family="ipv4", **kwargs):
    """
    .. versionadded:: 3002

    Ensure an nftables table is present

    name
        A user-defined table name.

    family
        Networking family, either ipv4 or ipv6
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    table_check = __salt__["nftables.check_table"](name, family=family)

    if table_check["result"] is True:
        ret["result"] = True
        ret["comment"] = "nftables table {} already exists in family {}".format(
            name, family
        )
        return ret

    if __opts__["test"]:
        ret["comment"] = "nftables table {} would be created in family {}".format(
            name, family
        )
        return ret

    res = __salt__["nftables.new_table"](name, family=family)

    if res["result"] is True:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "nftables table {} successfully created in family {}".format(
            name, family
        )
    else:
        ret["result"] = False
        ret["comment"] = f"Failed to create table {name} for family {family}"

    return ret


def table_absent(name, family="ipv4", **kwargs):
    """
    .. versionadded:: 3002

    Ensure an nftables table is absent

    name
        Name of the table to ensure is absent

    family
        Networking family, either ipv4 or ipv6
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    table_check = __salt__["nftables.check_table"](name, family)

    if table_check["result"] is False:
        ret["result"] = True
        ret["comment"] = "nftables table {} is already absent from family {}".format(
            name, family
        )
        return ret

    if __opts__["test"]:
        ret["comment"] = "nftables table {} would be deleted from family {}".format(
            name, family
        )
        return ret

    res = __salt__["nftables.delete_table"](name, family=family)

    if res["result"] is True:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "nftables table {} successfully deleted from family {}".format(
            name, family
        )
    else:
        ret["result"] = False
        ret["comment"] = f"Failed to delete table {name} from family {family}"

    return ret



"""
Pass Renderer for Salt
======================

pass_ is an encrypted on-disk password store.

.. _pass: https://www.passwordstore.org/

.. versionadded:: 2017.7.0

Setup
-----

*Note*: ``<user>`` needs to be replaced with the user salt-master will be
running as.

Have private gpg loaded into ``user``'s gpg keyring

.. code-block:: yaml

    load_private_gpg_key:
      cmd.run:
        - name: gpg --import <location_of_private_gpg_key>
        - unless: gpg --list-keys '<gpg_name>'

Said private key's public key should have been used when encrypting pass entries
that are of interest for pillar data.

Fetch and keep local pass git repo up-to-date

.. code-block:: yaml

        update_pass:
          git.latest:
            - force_reset: True
            - name: <git_repo>
            - target: /<user>/.password-store
            - identity: <location_of_ssh_private_key>
            - require:
              - cmd: load_private_gpg_key

Install pass binary

.. code-block:: yaml

        pass:
          pkg.installed

Salt master configuration options

.. code-block:: yaml

        # If the prefix is *not* set (default behavior), all template variables are
        # considered for fetching secrets from Pass. Those that cannot be resolved
        # to a secret are passed through.
        #
        # If the prefix is set, only the template variables with matching prefix are
        # considered for fetching the secrets, other variables are passed through.
        #
        # For ease of use it is recommended to set the following options as well:
        #       renderer: 'jinja|yaml|pass'
        #       pass_strict_fetch: true
        #
        pass_variable_prefix: 'pass:'

        # If set to 'true', error out when unable to fetch a secret for a template variable.
        pass_strict_fetch: true

        # Set GNUPGHOME env for Pass.
        # Defaults to: ~/.gnupg
        pass_gnupghome: <path>

        # Set PASSWORD_STORE_DIR env for Pass.
        # Defaults to: ~/.password-store
        pass_dir: <path>
"""

import logging
import os
from os.path import expanduser
from subprocess import PIPE, Popen

import salt.utils.path
from salt.exceptions import SaltConfigurationError, SaltRenderError

log = logging.getLogger(__name__)


def _get_pass_exec():
    """
    Return the pass executable or raise an error
    """
    pass_exec = salt.utils.path.which("pass")
    if pass_exec:
        return pass_exec
    else:
        raise SaltRenderError("pass unavailable")


def _fetch_secret(pass_path):
    """
    Fetch secret from pass based on pass_path. If there is
    any error, return back the original pass_path value
    """
    pass_exec = _get_pass_exec()

    # Make a backup in case we want to return the original value without stripped whitespaces
    original_pass_path = pass_path

    # Remove the optional prefix from pass path
    pass_prefix = __opts__["pass_variable_prefix"]
    if pass_prefix:
        # If we do not see our prefix we do not want to process this variable
        # and we return the unmodified pass path
        if not pass_path.startswith(pass_prefix):
            return pass_path

        # strip the prefix from the start of the string
        pass_path = pass_path[len(pass_prefix) :]

    # The pass_strict_fetch option must be used with pass_variable_prefix
    pass_strict_fetch = __opts__["pass_strict_fetch"]
    if pass_strict_fetch and not pass_prefix:
        msg = "The 'pass_strict_fetch' option requires 'pass_variable_prefix' option enabled"
        raise SaltConfigurationError(msg)

    # Remove whitespaces from the pass_path
    pass_path = pass_path.strip()

    cmd = [pass_exec, "show", pass_path]
    log.debug("Fetching secret: %s", " ".join(cmd))

    # Make sure environment variable HOME is set, since Pass looks for the
    # password-store under ~/.password-store.
    env = os.environ.copy()
    env["HOME"] = expanduser("~")

    pass_dir = __opts__["pass_dir"]
    if pass_dir:
        env["PASSWORD_STORE_DIR"] = pass_dir

    pass_gnupghome = __opts__["pass_gnupghome"]
    if pass_gnupghome:
        env["GNUPGHOME"] = pass_gnupghome

    try:
        proc = Popen(cmd, stdout=PIPE, stderr=PIPE, env=env, encoding="utf-8")
        pass_data, pass_error = proc.communicate()
        pass_returncode = proc.returncode
    except (OSError, UnicodeDecodeError) as e:
        pass_data, pass_error = "", str(e)
        pass_returncode = 1

    # The version of pass used during development sent output to
    # stdout instead of stderr even though its returncode was non zero.
    if pass_returncode or not pass_data:
        msg = f"Could not fetch secret '{pass_path}' from the password store: {pass_error}"
        if pass_strict_fetch:
            raise SaltRenderError(msg)
        else:
            log.warning(msg)
            return original_pass_path
    return pass_data.rstrip("\r\n")


def _decrypt_object(obj):
    """
    Recursively try to find a pass path (string) that can be handed off to pass
    """
    if isinstance(obj, str):
        return _fetch_secret(obj)
    elif isinstance(obj, dict):
        for pass_key, pass_path in obj.items():
            obj[pass_key] = _decrypt_object(pass_path)
    elif isinstance(obj, list):
        for pass_key, pass_path in enumerate(obj):
            obj[pass_key] = _decrypt_object(pass_path)
    return obj


def render(pass_info, saltenv="base", sls="", argline="", **kwargs):
    """
    Fetch secret from pass based on pass_path
    """
    return _decrypt_object(pass_info)



import logging

import pytest

from tests.support.helpers import dedent

log = logging.getLogger(__name__)


@pytest.mark.core_test
def test_issue_58763(tmp_path, modules, state_tree, caplog):

    sls_contents = dedent(
        """
    run_old:
      module.run:
        - name: test.random_hash
        - size: 10
        - hash_type: sha256

    run_new:
      module.run:
        - test.random_hash:
          - size: 10
          - hash_type: sha256
    """
    )
    with pytest.helpers.temp_file("issue-58763.sls", sls_contents, state_tree):
        with caplog.at_level(logging.DEBUG):
            ret = modules.state.sls(
                mods="issue-58763",
            )
            assert len(ret.raw) == 2
            for k in ret.raw:
                assert ret.raw[k]["result"] is True
            assert "Detected legacy module.run syntax: run_old" in caplog.messages
            assert "Using new style module.run syntax: run_new" in caplog.messages


@pytest.mark.core_test
def test_issue_58763_a(tmp_path, modules, state_tree, caplog):

    sls_contents = dedent(
        """
    test.random_hash:
      module.run:
        - size: 10
        - hash_type: sha256
    """
    )
    with pytest.helpers.temp_file("issue-58763.sls", sls_contents, state_tree):
        with caplog.at_level(logging.DEBUG):
            ret = modules.state.sls(
                mods="issue-58763",
            )
            assert len(ret.raw) == 1
            for k in ret.raw:
                assert ret.raw[k]["result"] is True
            assert (
                "Detected legacy module.run syntax: test.random_hash" in caplog.messages
            )


@pytest.mark.core_test
def test_issue_58763_b(tmp_path, modules, state_tree, caplog):

    sls_contents = dedent(
        """
    test.ping:
      module.run
    """
    )
    with pytest.helpers.temp_file("issue-58763.sls", sls_contents, state_tree):
        with caplog.at_level(logging.DEBUG):
            ret = modules.state.sls(
                mods="issue-58763",
            )
            assert len(ret.raw) == 1
            for k in ret.raw:
                assert ret.raw[k]["result"] is True
            assert "Detected legacy module.run syntax: test.ping" in caplog.messages


@pytest.mark.core_test
def test_issue_62988_a(tmp_path, modules, state_tree, caplog):

    sls_contents = dedent(
        """
    test_foo:
      test.succeed_with_changes

    run_new:
      module.wait:
        - test.random_hash:
          - size: 10
          - hash_type: sha256
        - watch:
          - test: test_foo
    """
    )
    with pytest.helpers.temp_file("issue-62988.sls", sls_contents, state_tree):
        with caplog.at_level(logging.DEBUG):
            ret = modules.state.sls(
                mods="issue-62988",
            )
            assert len(ret.raw) == 2
            for k in ret.raw:
                assert ret.raw[k]["result"] is True
            assert "Using new style module.run syntax: run_new" in caplog.messages


@pytest.mark.core_test
def test_issue_62988_b(tmp_path, modules, state_tree, caplog):

    sls_contents = dedent(
        """
    test_foo:
      test.succeed_with_changes:
        - watch_in:
          - module: run_new

    run_new:
      module.wait:
        - test.random_hash:
          - size: 10
          - hash_type: sha256
    """
    )
    with pytest.helpers.temp_file("issue-62988.sls", sls_contents, state_tree):
        with caplog.at_level(logging.DEBUG):
            ret = modules.state.sls(
                mods="issue-62988",
            )
            assert len(ret.raw) == 2
            for k in ret.raw:
                assert ret.raw[k]["result"] is True
            assert "Using new style module.run syntax: run_new" in caplog.messages



"""
    :codeauthor: Gareth J. Greenaway <ggreenaway@vmware.com>
"""

import pytest

import salt.modules.mongodb as mongodb
from tests.support.mock import MagicMock, patch

try:
    import pymongo  # pylint: disable=unused-import

    HAS_PYMONGO = True
except ImportError:
    HAS_PYMONGO = False


pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skipif(not HAS_PYMONGO, reason="No python mongo client installed."),
]


class MockMongoConnect:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def list_database_names(self, *args, **kwards):
        return []

    def drop_database(self, *args, **kwards):
        return True


class MockInsertResult:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        self.inserted_ids = []
        self.acknowledged = True


class MockDeleteResult:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        self.deleted_count = 0
        self.raw_result = {}
        self.acknowledged = True


class MockPyMongoDatabase:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        self.my_collection = None

    def authenticate(self, *args, **kwards):
        return True

    def command(self, *args, **kwards):
        return ""

    def create_collection(self, *args, **kwards):
        return True

    def list_collection_names(self, *args, **kwards):
        return []


class MockPyMongoCollection:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def find(self, *args, **kwards):
        return []

    def insert_many(self, *args, **kwards):
        return True

    def delete_one(self, *args, **kwards):
        return True


@pytest.fixture
def configure_loader_modules():
    return {mongodb: {}}


def test_version():
    """
    Test mongodb.version
    """
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    database_command_mock = MagicMock(autospec=True, return_value={"version": "6.0.2"})
    config_option_mock = MagicMock(
        side_effect=["user", "password", "localhost", "27017"]
    )

    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", database_command_mock
    )
    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_command, patch_pymongo_database, patch_salt_dict:
        ret = mongodb.version()
        assert ret == "6.0.2"


def test_db_list():
    """
    Test mongodb.db_list
    """
    list_db_names_mock = MagicMock(
        autospec=True, return_value=["admin", "config", "local"]
    )
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())

    patch_list_db_names = patch.object(
        MockMongoConnect, "list_database_names", list_db_names_mock
    )
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_list_db_names
    ), patch_mongo_client, patch_pymongo_database, patch_salt_dict:
        ret = mongodb.db_list()
        assert ret == ["admin", "config", "local"]


def test_db_exists():
    """
    Test mongodb.db_exists
    """
    list_db_names_mock = MagicMock(
        autospec=True, return_value=["admin", "config", "local"]
    )
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_list_db_names = patch.object(
        MockMongoConnect, "list_database_names", list_db_names_mock
    )
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_list_db_names
    ), patch_mongo_client, patch_pymongo_database, patch_salt_dict:
        ret = mongodb.db_exists("admin")
        assert ret


def test_user_list():
    """
    Test mongodb.user_list
    """
    user_info = {
        "users": [
            {
                "_id": "admin.test_user",
                "userId": "",
                "user": "test_user",
                "db": "admin",
                "roles": [{"role": "read", "db": "admin"}],
                "mechanisms": ["SCRAM-SHA-1", "SCRAM-SHA-256"],
            },
            {
                "_id": "admin.test_user2",
                "userId": "",
                "user": "test_user2",
                "db": "admin",
                "roles": [],
                "mechanisms": ["SCRAM-SHA-1", "SCRAM-SHA-256"],
            },
        ],
        "ok": 1.0,
    }
    database_version = {"version": "6.0.2"}

    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    database_command_mock = MagicMock(side_effect=[database_version, user_info])
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", database_command_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_pymongo_command, patch_salt_dict:
        ret = mongodb.user_list()
        expected = [
            {
                "user": "test_user",
                "roles": [{"role": "read", "db": "admin"}],
            },
            {"user": "test_user2", "roles": []},
        ]
        assert ret == expected


def test_user_exists():
    """
    Test mongodb.user_exists
    """
    user_info = {
        "users": [
            {
                "_id": "admin.test_user",
                "userId": "",
                "user": "test_user",
                "db": "admin",
                "roles": [{"role": "read", "db": "admin"}],
                "mechanisms": ["SCRAM-SHA-1", "SCRAM-SHA-256"],
            },
            {
                "_id": "admin.test_user2",
                "userId": "",
                "user": "test_user2",
                "db": "admin",
                "roles": [],
                "mechanisms": ["SCRAM-SHA-1", "SCRAM-SHA-256"],
            },
        ],
        "ok": 1.0,
    }
    database_version = {"version": "6.0.2"}

    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)

    with patch_mongo_client, patch_pymongo_database:
        config_option_mock = MagicMock(
            autospec=True, side_effect=["user", "password", "localhost", "27017"]
        )
        database_command_mock = MagicMock(
            autospec=True, side_effect=[database_version, user_info]
        )

        patch_salt_dict = patch.dict(
            mongodb.__salt__, {"config.option": config_option_mock}
        )
        patch_pymongo_command = patch.object(
            MockPyMongoDatabase, "command", database_command_mock
        )

        with patch_salt_dict, patch_pymongo_command:
            ret = mongodb.user_exists("test_user")
            assert ret

        config_option_mock = MagicMock(
            autospec=True, side_effect=["user", "password", "localhost", "27017"]
        )
        database_command_mock = MagicMock(
            autospec=True, side_effect=[database_version, user_info]
        )

        patch_salt_dict = patch.dict(
            mongodb.__salt__, {"config.option": config_option_mock}
        )
        patch_pymongo_command = patch.object(
            MockPyMongoDatabase, "command", database_command_mock
        )

        with patch_salt_dict, patch_pymongo_command:
            ret = mongodb.user_exists("no_test_user")
            assert not ret


def test_user_create():
    """
    Test mongodb.user_create
    """
    user_create_mock = MagicMock(return_value={"ok": 1.0})

    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", user_create_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_command:
        ret = mongodb.user_create("test_user", "test_password")
        assert ret


def test_user_create_exception():
    """
    Test mongodb.user_create
    """
    user_create_mock = MagicMock(side_effect=pymongo.errors.PyMongoError)

    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", user_create_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_command:
        ret = mongodb.user_create("test_user", "test_password")
        assert not ret


def test_user_remove():
    """
    Test mongodb.user_remove
    """
    user_remove_mock = MagicMock(autospec=True, return_value={"ok": 1.0})

    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", user_remove_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_command:
        ret = mongodb.user_remove("test_user")
        assert ret


def test_user_remove_exception():
    """
    Test mongodb.user_remove
    """
    user_remove_mock = MagicMock(autospec=True, side_effect=pymongo.errors.PyMongoError)

    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", user_remove_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_command:
        ret = mongodb.user_remove("test_user")
        assert not ret


def test_user_roles_exists():
    """
    Test mongodb.user_roles_exist
    """
    user_info = {
        "users": [
            {
                "_id": "admin.test_user",
                "userId": "",
                "user": "test_user",
                "db": "admin",
                "roles": [{"role": "read", "db": "admin"}],
                "mechanisms": ["SCRAM-SHA-1", "SCRAM-SHA-256"],
            },
            {
                "_id": "admin.test_user2",
                "userId": "",
                "user": "test_user2",
                "db": "admin",
                "roles": [],
                "mechanisms": ["SCRAM-SHA-1", "SCRAM-SHA-256"],
            },
        ],
        "ok": 1.0,
    }
    database_version = {"version": "6.0.2"}

    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )
    database_command_mock = MagicMock(
        autospec=True, side_effect=[database_version, user_info]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", database_command_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_command:
        ret = mongodb.user_roles_exists("test_user", '["read"]', "admin")
        assert ret


def test_user_grant_roles():
    """
    Test mongodb.user_remove
    """
    user_grant_roles_mock = MagicMock(autospec=True, return_value={"ok": 1.0})
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", user_grant_roles_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_command:
        ret = mongodb.user_grant_roles(
            "test_user", '[{"role": "readWrite", "db": "admin" }]', "admin"
        )
        assert ret


def test_user_revoke_roles():
    """
    Test mongodb.user_remove
    """
    user_revoke_roles_mock = MagicMock(autospec=True, return_value={"ok": 1.0})
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", user_revoke_roles_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_command:
        ret = mongodb.user_revoke_roles(
            "test_user", '[{"role": "readWrite", "db": "admin" }]', "admin"
        )
        assert ret


def test_collection_create():
    """
    Test mongodb.user_create
    """
    collection_create_mock = MagicMock(autospec=True, return_value={"ok": 1.0})
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_command = patch.object(
        MockPyMongoDatabase, "command", collection_create_mock
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_command:
        ret = mongodb.collection_create("test_collection")
        assert ret


def test_collections_list():
    """
    Test mongodb.collections_list
    """
    collections_list = MagicMock(
        autospec=True, return_value=["system.users", "mycollection", "system.version"]
    )
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )

    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_list_collection_names = patch.object(
        MockPyMongoDatabase, "list_collection_names", collections_list
    )
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )

    with (
        patch_mongo_client
    ), patch_pymongo_database, patch_salt_dict, patch_pymongo_list_collection_names:
        ret = mongodb.collections_list()
        assert ret == ["system.users", "mycollection", "system.version"]


def test_insert():
    """
    Test mongodb.insert
    """
    collection_insert_mock = MockInsertResult()

    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    config_option_mock = MagicMock(
        autospec=True, side_effect=["user", "password", "localhost", "27017"]
    )
    pymongo_collection_mock = MagicMock(
        autospec=True, return_value=MockPyMongoCollection()
    )

    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_salt_dict = patch.dict(
        mongodb.__salt__, {"config.option": config_option_mock}
    )
    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_pymongo_collection = patch.object(mongodb, "getattr", pymongo_collection_mock)

    with (
        patch_mongo_client
    ), patch_salt_dict, patch_pymongo_database, patch_pymongo_collection:
        patch_pymongo_collection_insert = patch.object(
            MockPyMongoCollection,
            "insert_many",
            MagicMock(return_value=collection_insert_mock),
        )
        with patch_pymongo_collection_insert:
            ret = mongodb.insert(
                '[{"foo": "FOO", "bar": "BAR"}, {"foo": "BAZ", "bar": "BAM"}]',
                "my_collection",
            )
            assert ret


def test_find():
    """
    Test mongodb.find
    """
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    pymongo_collection_mock = MagicMock(
        autospec=True, return_value=MockPyMongoCollection()
    )

    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_pymongo_collection = patch.object(mongodb, "getattr", pymongo_collection_mock)

    with patch_mongo_client, patch_pymongo_database, patch_pymongo_collection:
        collection_find_mock = [
            {"_id": "63459d7f78548d1d02295dd0", "foo": "FOO", "bar": "BAR"},
            {"_id": "6345c4fea9a1255a430b2fef", "foo": "FOO", "bar": "BAR"},
            {"_id": "6345c505961560271e34a22a", "foo": "FOO", "bar": "BAR"},
        ]

        config_option_mock = MagicMock(
            autospec=True, side_effect=["user", "password", "localhost", "27017"]
        )
        patch_salt_dict = patch.dict(
            mongodb.__salt__, {"config.option": config_option_mock}
        )
        patch_pymongo_collection_find = patch.object(
            MockPyMongoCollection, "find", MagicMock(return_value=collection_find_mock)
        )
        with patch_pymongo_collection_find, patch_salt_dict:
            expected = [
                {"_id": "63459d7f78548d1d02295dd0", "foo": "FOO", "bar": "BAR"},
                {"_id": "6345c4fea9a1255a430b2fef", "foo": "FOO", "bar": "BAR"},
                {"_id": "6345c505961560271e34a22a", "foo": "FOO", "bar": "BAR"},
                {"_id": "63459d7f78548d1d02295dd0", "foo": "FOO", "bar": "BAR"},
                {"_id": "6345c4fea9a1255a430b2fef", "foo": "FOO", "bar": "BAR"},
                {"_id": "6345c505961560271e34a22a", "foo": "FOO", "bar": "BAR"},
            ]
            ret = mongodb.find(
                "test_collection",
                ['{"foo": "FOO", "bar": "BAR"}', '{"foo": "BAZ", "bar": "BAM"}'],
            )
            assert ret == expected

        collection_find_mock = [{"_id": "63459d7f78548d1d02295dd0", "baz": "BAZ"}]

        config_option_mock = MagicMock(
            autospec=True, side_effect=["user", "password", "localhost", "27017"]
        )
        patch_salt_dict = patch.dict(
            mongodb.__salt__, {"config.option": config_option_mock}
        )
        patch_pymongo_collection_find = patch.object(
            MockPyMongoCollection, "find", MagicMock(return_value=collection_find_mock)
        )
        with patch_pymongo_collection_find, patch_salt_dict:
            expected = [{"_id": "63459d7f78548d1d02295dd0", "baz": "BAZ"}]
            ret = mongodb.find("my_collection", {"baz": "BAZ"})
            assert ret == expected


def test_remove():
    """
    Test mongodb.remove
    """
    mongodb_client_mock = MagicMock(autospec=True, return_value=MockMongoConnect())
    pymongo_database_mock = MagicMock(autospec=True, return_value=MockPyMongoDatabase())
    pymongo_collection_mock = MagicMock(
        autospec=True, return_value=MockPyMongoCollection()
    )

    patch_mongo_client = patch("pymongo.MongoClient", mongodb_client_mock)
    patch_pymongo_database = patch("pymongo.database.Database", pymongo_database_mock)
    patch_pymongo_collection = patch.object(mongodb, "getattr", pymongo_collection_mock)

    with patch_mongo_client, patch_pymongo_database, patch_pymongo_collection:
        config_option_mock = MagicMock(
            autospec=True, side_effect=["user", "password", "localhost", "27017"]
        )
        patch_salt_dict = patch.dict(
            mongodb.__salt__, {"config.option": config_option_mock}
        )

        # Assume we delete one entry each time
        collection_delete_one_mock = MockDeleteResult()
        collection_delete_one_mock.deleted_count = 1
        collection_delete_one_mock.raw_result = {"n": 1, "ok": 1.0}
        collection_delete_one_mock.acknowledged = True

        patch_pymongo_collection_remove = patch.object(
            MockPyMongoCollection,
            "delete_one",
            MagicMock(return_value=collection_delete_one_mock),
        )
        with patch_pymongo_collection_remove, patch_salt_dict:

            ret = mongodb.remove(
                "test_collection",
                ['{"foo": "FOO", "bar": "BAR"}', '{"foo": "BAZ", "bar": "BAM"}'],
            )
            expected = "2 objects removed"
            assert ret == expected

        config_option_mock = MagicMock(
            autospec=True, side_effect=["user", "password", "localhost", "27017"]
        )
        patch_salt_dict = patch.dict(
            mongodb.__salt__, {"config.option": config_option_mock}
        )

        # Assume we delete one entry each time
        collection_delete_one_mock = MockDeleteResult()
        collection_delete_one_mock.deleted_count = 1
        collection_delete_one_mock.raw_result = {"n": 1, "ok": 1.0}
        collection_delete_one_mock.acknowledged = True

        patch_pymongo_collection_remove = patch.object(
            MockPyMongoCollection,
            "delete_one",
            MagicMock(return_value=collection_delete_one_mock),
        )
        with patch_pymongo_collection_remove, patch_salt_dict:

            ret = mongodb.remove("test_collection", {"foo": "FOO", "bar": "BAR"})
            expected = "1 objects removed"
            assert ret == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.augeas_cfg
"""

import pytest

import salt.modules.augeas_cfg as augeas_cfg
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch

# Make sure augeas python interface is installed
if augeas_cfg.HAS_AUGEAS:
    from augeas import Augeas as _Augeas


pytestmark = [
    pytest.mark.skipif(
        augeas_cfg.HAS_AUGEAS is False,
        reason="python-augeas is required for this test case",
    )
]


@pytest.fixture
def configure_loader_modules():
    return {augeas_cfg: {}}


def test_execute():
    """
    Test if it execute Augeas commands
    """
    assert augeas_cfg.execute() == {"retval": True}


def test_execute_io_error():
    """
    Test if it execute Augeas commands
    """
    ret = {"error": "Command  is not supported (yet)", "retval": False}
    assert augeas_cfg.execute(None, None, [" "]) == ret


def test_execute_value_error():
    """
    Test if it execute Augeas commands
    """
    ret = {
        "retval": False,
        "error": "Invalid formatted command, see debug log for details: ",
    }
    assert augeas_cfg.execute(None, None, ["set "]) == ret


# 'get' function tests: 1


def test_get():
    """
    Test if it get a value for a specific augeas path
    """
    mock = MagicMock(side_effect=RuntimeError("error"))
    with patch.object(_Augeas, "match", mock):
        assert augeas_cfg.get("/etc/hosts") == {"error": "error"}

    mock = MagicMock(return_value=True)
    with patch.object(_Augeas, "match", mock):
        assert augeas_cfg.get("/etc/hosts") == {"/etc/hosts": None}


# 'setvalue' function tests: 4


def test_setvalue():
    """
    Test if it set a value for a specific augeas path
    """
    assert augeas_cfg.setvalue("prefix=/etc/hosts") == {"retval": True}


def test_setvalue_io_error():
    """
    Test if it set a value for a specific augeas path
    """
    mock = MagicMock(side_effect=IOError(""))
    with patch.object(_Augeas, "save", mock):
        assert augeas_cfg.setvalue("prefix=/files/etc/") == {
            "retval": False,
            "error": "",
        }


def test_setvalue_uneven_path():
    """
    Test if it set a value for a specific augeas path
    """
    mock = MagicMock(side_effect=RuntimeError("error"))
    with patch.object(_Augeas, "match", mock):
        pytest.raises(
            SaltInvocationError,
            augeas_cfg.setvalue,
            ["/files/etc/hosts/1/canonical", "localhost"],
        )


def test_setvalue_one_prefix():
    """
    Test if it set a value for a specific augeas path
    """
    pytest.raises(
        SaltInvocationError,
        augeas_cfg.setvalue,
        "prefix=/files",
        "10.18.1.1",
        "prefix=/etc",
        "test",
    )


# 'match' function tests: 2


def test_match():
    """
    Test if it matches for path expression
    """
    assert augeas_cfg.match("/etc/service", "ssh") == {}


def test_match_runtime_error():
    """
    Test if it matches for path expression
    """
    mock = MagicMock(side_effect=RuntimeError("error"))
    with patch.object(_Augeas, "match", mock):
        assert augeas_cfg.match("/etc/service-name", "ssh") == {}


# 'remove' function tests: 2


def test_remove():
    """
    Test if it removes for path expression
    """
    assert augeas_cfg.remove("/etc/service") == {"count": 0, "retval": True}


def test_remove_io_runtime_error():
    """
    Test if it removes for path expression
    """
    mock = MagicMock(side_effect=RuntimeError("error"))
    with patch.object(_Augeas, "save", mock):
        assert augeas_cfg.remove("/etc/service-name") == {
            "count": 0,
            "error": "error",
            "retval": False,
        }


# 'ls' function tests: 1


def test_ls():
    """
    Test if it list the direct children of a node
    """
    assert augeas_cfg.ls("/etc/passwd") == {}


# 'tree' function tests: 1


def test_tree():
    """
    Test if it returns recursively the complete tree of a node
    """
    assert augeas_cfg.tree("/etc/") == {"/etc": None}



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import os
import textwrap

import pytest

import salt.modules.dnsmasq as dnsmasq
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {dnsmasq: {}}


def test_version():
    """
    test to show installed version of dnsmasq.
    """
    mock = MagicMock(return_value="A B C")
    with patch.dict(dnsmasq.__salt__, {"cmd.run": mock}):
        assert dnsmasq.version() == "C"


def test_fullversion():
    """
    Test to Show installed version of dnsmasq and compile options.
    """
    mock = MagicMock(return_value="A B C\nD E F G H I")
    with patch.dict(dnsmasq.__salt__, {"cmd.run": mock}):
        assert dnsmasq.fullversion() == {
            "version": "C",
            "compile options": ["G", "H", "I"],
        }


def test_set_config():
    """
    test to show installed version of dnsmasq.
    """
    mock = MagicMock(return_value={"conf-dir": "A"})
    with patch.object(dnsmasq, "get_config", mock):
        mock = MagicMock(return_value=[".", "~", "bak", "#"])
        with patch.object(os, "listdir", mock):
            assert dnsmasq.set_config() == {}


def test_set_config_filter_pub_kwargs():
    """
    Test that the kwargs returned from running the set_config function
    do not contain the __pub that may have been passed through in **kwargs.
    """
    with patch(
        "salt.modules.dnsmasq.get_config", MagicMock(return_value={"conf-dir": "A"})
    ):
        mock_domain = "local"
        mock_address = "/some-test-address.local/8.8.4.4"
        with patch.dict(dnsmasq.__salt__, {"file.append": MagicMock()}):
            ret = dnsmasq.set_config(
                follow=False,
                domain=mock_domain,
                address=mock_address,
                __pub_pid=8184,
                __pub_jid=20161101194639387946,
                __pub_tgt="salt-call",
            )
        assert ret == {"domain": mock_domain, "address": mock_address}


def test_get_config():
    """
    test to dumps all options from the config file.
    """
    mock = MagicMock(return_value={"conf-dir": "A"})
    with patch.object(dnsmasq, "get_config", mock):
        mock = MagicMock(return_value=[".", "~", "bak", "#"])
        with patch.object(os, "listdir", mock):
            assert dnsmasq.get_config() == {"conf-dir": "A"}


def test_parse_dnsmasq_no_file():
    """
    Tests that a CommandExecutionError is when a filename that doesn't exist is
    passed in.
    """
    pytest.raises(CommandExecutionError, dnsmasq._parse_dnamasq, "filename")


def test_parse_dnamasq():
    """
    test for generic function for parsing dnsmasq files including includes.
    """
    with patch("os.path.isfile", MagicMock(return_value=True)):
        text_file_data = textwrap.dedent(
            """\
            line here
            second line
            A=B
            #"""
        )
        with patch("salt.utils.files.fopen", mock_open(read_data=text_file_data)):
            assert dnsmasq._parse_dnamasq("filename") == {
                "A": "B",
                "unparsed": ["line here\n", "second line\n"],
            }



"""
Common logic used by the docker state and execution module

This module contains logic to accommodate docker/salt CLI usage, as well as
input as formatted by states.
"""

import copy
import logging

import salt.utils.args
import salt.utils.data
import salt.utils.dockermod.translate
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.utils.args import get_function_argspec as _argspec
from salt.utils.dockermod.translate.helpers import split as _split

try:
    import docker
except ImportError:
    docker = None

# These next two imports are only necessary to have access to the needed
# functions so that we can get argspecs for the container config, host config,
# and networking config (see the get_client_args() function).
try:
    import docker.types  # pylint: disable=no-name-in-module
except ImportError:
    pass
try:
    import docker.utils  # pylint: disable=no-name-in-module
except ImportError:
    pass

NOTSET = object()

__virtualname__ = "docker"

# Default timeout as of docker-py 1.0.0
CLIENT_TIMEOUT = 60
# Timeout for stopping the container, before a kill is invoked
SHUTDOWN_TIMEOUT = 10

log = logging.getLogger(__name__)


def __virtual__():
    if docker is None:
        return False
    return __virtualname__


def get_client_args(limit=None):
    if docker is None:
        raise CommandExecutionError("docker Python module not imported")

    limit = salt.utils.args.split_input(limit or [])
    ret = {}

    if not limit or any(
        x in limit
        for x in ("create_container", "host_config", "connect_container_to_network")
    ):
        try:
            ret["create_container"] = _argspec(docker.APIClient.create_container).args
        except AttributeError:
            try:
                ret["create_container"] = _argspec(docker.Client.create_container).args
            except AttributeError:
                raise CommandExecutionError("Coult not get create_container argspec")

        try:
            ret["host_config"] = _argspec(docker.types.HostConfig.__init__).args
        except AttributeError:
            try:
                ret["host_config"] = _argspec(docker.utils.create_host_config).args
            except AttributeError:
                raise CommandExecutionError("Could not get create_host_config argspec")

        try:
            ret["connect_container_to_network"] = _argspec(
                docker.types.EndpointConfig.__init__
            ).args
        except AttributeError:
            try:
                ret["connect_container_to_network"] = _argspec(
                    docker.utils.utils.create_endpoint_config
                ).args
            except AttributeError:
                try:
                    ret["connect_container_to_network"] = _argspec(
                        docker.utils.create_endpoint_config
                    ).args
                except AttributeError:
                    raise CommandExecutionError(
                        "Could not get connect_container_to_network argspec"
                    )

    for key, wrapped_func in (
        ("logs", docker.api.container.ContainerApiMixin.logs),
        ("create_network", docker.api.network.NetworkApiMixin.create_network),
    ):
        if not limit or key in limit:
            try:
                func_ref = wrapped_func
                try:
                    # functools.wraps makes things a little easier in Python 3
                    ret[key] = _argspec(func_ref.__wrapped__).args
                except AttributeError:
                    # functools.wraps changed (unlikely), bail out
                    ret[key] = []
            except AttributeError:
                # Function moved, bail out
                ret[key] = []

    if not limit or "ipam_config" in limit:
        try:
            ret["ipam_config"] = _argspec(docker.types.IPAMPool.__init__).args
        except AttributeError:
            try:
                ret["ipam_config"] = _argspec(docker.utils.create_ipam_pool).args
            except AttributeError:
                raise CommandExecutionError("Could not get ipam args")

    for item in ret:
        # The API version is passed automagically by the API code that imports
        # these classes/functions and is not an arg that we will be passing, so
        # remove it if present. Similarly, don't include "self" if it shows up
        # in the arglist.
        for argname in ("version", "self"):
            try:
                ret[item].remove(argname)
            except ValueError:
                pass

    # Remove any args in host or endpoint config from the create_container
    # arglist. This keeps us from accidentally allowing args that docker-py has
    # moved from the create_container function to the either the host or
    # endpoint config.
    for item in ("host_config", "connect_container_to_network"):
        for val in ret.get(item, []):
            try:
                ret["create_container"].remove(val)
            except ValueError:
                # Arg is not in create_container arglist
                pass

    for item in ("create_container", "host_config", "connect_container_to_network"):
        if limit and item not in limit:
            ret.pop(item, None)

    try:
        ret["logs"].remove("container")
    except (KeyError, ValueError, TypeError):
        pass

    return ret


def translate_input(
    translator,
    skip_translate=None,
    ignore_collisions=False,
    validate_ip_addrs=True,
    **kwargs
):
    """
    Translate CLI/SLS input into the format the API expects. The ``translator``
    argument must be a module containing translation functions, within
    salt.utils.dockermod.translate. A ``skip_translate`` kwarg can be passed to
    control which arguments are translated. It can be either a comma-separated
    list or an iterable containing strings (e.g. a list or tuple), and members
    of that tuple will have their translation skipped. Optionally,
    skip_translate can be set to True to skip *all* translation.
    """
    kwargs = copy.deepcopy(salt.utils.args.clean_kwargs(**kwargs))
    invalid = {}
    collisions = []

    if skip_translate is True:
        # Skip all translation
        return kwargs
    else:
        if not skip_translate:
            skip_translate = ()
        else:
            try:
                skip_translate = _split(skip_translate)
            except AttributeError:
                pass
            if not hasattr(skip_translate, "__iter__"):
                log.error("skip_translate is not an iterable, ignoring")
                skip_translate = ()

    try:
        # Using list(kwargs) here because if there are any invalid arguments we
        # will be popping them from the kwargs.
        for key in list(kwargs):
            real_key = translator.ALIASES.get(key, key)
            if real_key in skip_translate:
                continue

            # ipam_pools is designed to be passed as a list of actual
            # dictionaries, but if each of the dictionaries passed has a single
            # element, it will be incorrectly repacked.
            if key != "ipam_pools" and salt.utils.data.is_dictlist(kwargs[key]):
                kwargs[key] = salt.utils.data.repack_dictlist(kwargs[key])

            try:
                kwargs[key] = getattr(translator, real_key)(
                    kwargs[key],
                    validate_ip_addrs=validate_ip_addrs,
                    skip_translate=skip_translate,
                )
            except AttributeError:
                log.debug("No translation function for argument '%s'", key)
                continue
            except SaltInvocationError as exc:
                kwargs.pop(key)
                invalid[key] = exc.strerror

        try:
            translator._merge_keys(kwargs)
        except AttributeError:
            pass

        # Convert CLI versions of commands to their docker-py counterparts
        for key in translator.ALIASES:
            if key in kwargs:
                new_key = translator.ALIASES[key]
                value = kwargs.pop(key)
                if new_key in kwargs:
                    collisions.append(new_key)
                else:
                    kwargs[new_key] = value

        try:
            translator._post_processing(kwargs, skip_translate, invalid)
        except AttributeError:
            pass

    except Exception as exc:  # pylint: disable=broad-except
        error_message = str(exc)
        log.error("Error translating input: '%s'", error_message, exc_info=True)
    else:
        error_message = None

    error_data = {}
    if error_message is not None:
        error_data["error_message"] = error_message
    if invalid:
        error_data["invalid"] = invalid
    if collisions and not ignore_collisions:
        for item in collisions:
            error_data.setdefault("collisions", []).append(
                "'{}' is an alias for '{}', they cannot both be used".format(
                    translator.ALIASES_REVMAP[item], item
                )
            )
    if error_data:
        raise CommandExecutionError("Failed to translate input", info=error_data)

    return kwargs


def create_ipam_config(*pools, **kwargs):
    """
    Builds an IP address management (IPAM) config dictionary
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)

    try:
        # docker-py 2.0 and newer
        pool_args = salt.utils.args.get_function_argspec(
            docker.types.IPAMPool.__init__
        ).args
        create_pool = docker.types.IPAMPool
        create_config = docker.types.IPAMConfig
    except AttributeError:
        # docker-py < 2.0
        pool_args = salt.utils.args.get_function_argspec(
            docker.utils.create_ipam_pool
        ).args
        create_pool = docker.utils.create_ipam_pool
        create_config = docker.utils.create_ipam_config

    for primary_key, alias_key in (("driver", "ipam_driver"), ("options", "ipam_opts")):

        if alias_key in kwargs:
            alias_val = kwargs.pop(alias_key)
            if primary_key in kwargs:
                log.warning(
                    "docker.create_ipam_config: Both '%s' and '%s' "
                    "passed. Ignoring '%s'",
                    alias_key,
                    primary_key,
                    alias_key,
                )
            else:
                kwargs[primary_key] = alias_val

    if salt.utils.data.is_dictlist(kwargs.get("options")):
        kwargs["options"] = salt.utils.data.repack_dictlist(kwargs["options"])

    # Get all of the IPAM pool args that were passed as individual kwargs
    # instead of in the *pools tuple
    pool_kwargs = {}
    for key in list(kwargs):
        if key in pool_args:
            pool_kwargs[key] = kwargs.pop(key)

    pool_configs = []
    if pool_kwargs:
        pool_configs.append(create_pool(**pool_kwargs))
    pool_configs.extend([create_pool(**pool) for pool in pools])

    if pool_configs:
        # Sanity check the IPAM pools. docker-py's type/function for creating
        # an IPAM pool will allow you to create a pool with a gateway, IP
        # range, or map of aux addresses, even when no subnet is passed.
        # However, attempting to use this IPAM pool when creating the network
        # will cause the Docker Engine to throw an error.
        if any("Subnet" not in pool for pool in pool_configs):
            raise SaltInvocationError("A subnet is required in each IPAM pool")
        else:
            kwargs["pool_configs"] = pool_configs

    ret = create_config(**kwargs)
    pool_dicts = ret.get("Config")
    if pool_dicts:
        # When you inspect a network with custom IPAM configuration, only
        # arguments which were explictly passed are reflected. By contrast,
        # docker-py will include keys for arguments which were not passed in
        # but set the value to None. Thus, for ease of comparison, the below
        # loop will remove all keys with a value of None from the generated
        # pool configs.
        for idx, _ in enumerate(pool_dicts):
            for key in list(pool_dicts[idx]):
                if pool_dicts[idx][key] is None:
                    del pool_dicts[idx][key]

    return ret



"""
Mounting of filesystems
=======================

Mount any type of mountable filesystem with the mounted function:

.. code-block:: yaml

    /mnt/sdb:
      mount.mounted:
        - device: /dev/sdb1
        - fstype: ext4
        - mkmnt: True
        - opts:
          - defaults

    /srv/bigdata:
      mount.mounted:
        - device: UUID=066e0200-2867-4ebe-b9e6-f30026ca2314
        - fstype: xfs
        - opts: nobootwait,noatime,nodiratime,nobarrier,logbufs=8
        - dump: 0
        - pass_num: 2
        - persist: True
        - mkmnt: True

    /var/lib/bigdata:
      mount.mounted:
        - device: /srv/bigdata
        - fstype: none
        - opts: bind
        - dump: 0
        - pass_num: 0
        - persist: True
        - mkmnt: True
"""

import logging
import os.path
import re

log = logging.getLogger(__name__)


def _size_convert(_re_size):
    converted_size = int(_re_size.group("size_value"))
    if _re_size.group("size_unit") == "m":
        converted_size = int(converted_size) * 1024
    if _re_size.group("size_unit") == "g":
        converted_size = int(converted_size) * 1024 * 1024
    return converted_size


def mounted(
    name,
    device,
    fstype,
    mkmnt=False,
    opts="defaults",
    dump=0,
    pass_num=0,
    config="/etc/fstab",
    persist=True,
    mount=True,
    user=None,
    match_on="auto",
    device_name_regex=None,
    extra_mount_invisible_options=None,
    extra_mount_invisible_keys=None,
    extra_mount_ignore_fs_keys=None,
    extra_mount_translate_options=None,
    hidden_opts=None,
    bind_mount_copy_active_opts=True,
    **kwargs,
):
    """
    Verify that a device is mounted

    name
        The path to the location where the device is to be mounted

    device
        The device name, typically the device node, such as ``/dev/sdb1``
        or ``UUID=066e0200-2867-4ebe-b9e6-f30026ca2314`` or ``LABEL=DATA``

    fstype
        The filesystem type, this will be ``xfs``, ``ext2/3/4`` in the case of classic
        filesystems, ``fuse`` in the case of fuse mounts, and ``nfs`` in the case of nfs mounts

    mkmnt
        If the mount point is not present then the state will fail, set ``mkmnt: True``
        to create the mount point if it is otherwise not present

    opts
        A list object of options or a comma delimited list

    dump
        The dump value to be passed into the fstab, Default is ``0``

    pass_num
        The pass value to be passed into the fstab, Default is ``0``

    config
        Set an alternative location for the fstab, Default is ``/etc/fstab``

    persist
        Set if the mount should be saved in the fstab, Default is ``True``

    mount
        Set if the mount should be mounted immediately, Default is ``True``

    user
        The account used to execute the mount; this defaults to the user salt is
        running as on the minion

    match_on
        A name or list of fstab properties on which this state should be applied.
        Default is ``auto``, a special value indicating to guess based on fstype.
        In general, ``auto`` matches on name for recognized special devices and
        device otherwise.

    device_name_regex
        A list of device exact names or regular expressions which should
        not force a remount. For example, glusterfs may be mounted with a
        comma-separated list of servers in fstab, but the /proc/self/mountinfo
        will show only the first available server.

        .. code-block:: jinja

            {% set glusterfs_ip_list = ['10.0.0.1', '10.0.0.2', '10.0.0.3'] %}

            mount glusterfs volume:
              mount.mounted:
                - name: /mnt/glusterfs_mount_point
                - device: {{ glusterfs_ip_list|join(',') }}:/volume_name
                - fstype: glusterfs
                - opts: _netdev,rw,defaults,direct-io-mode=disable
                - mkmnt: True
                - persist: True
                - dump: 0
                - pass_num: 0
                - device_name_regex:
                  - ({{ glusterfs_ip_list|join('|') }}):/volume_name

        .. versionadded:: 2016.11.0

    extra_mount_invisible_options
        A list of extra options that are not visible through the
        ``/proc/self/mountinfo`` interface.

        If a option is not visible through this interface it will always remount
        the device. This option extends the builtin ``mount_invisible_options``
        list.

    extra_mount_invisible_keys
        A list of extra key options that are not visible through the
        ``/proc/self/mountinfo`` interface.

        If a key option is not visible through this interface it will always
        remount the device. This option extends the builtin
        ``mount_invisible_keys`` list.

        A good example for a key option is the password option::

            password=badsecret

    extra_mount_ignore_fs_keys
        A dict of filesystem options which should not force a remount. This will update
        the internal dictionary. The dict should look like this::

            {
                'ramfs': ['size']
            }

    extra_mount_translate_options
        A dict of mount options that gets translated when mounted. To prevent a remount
        add additional options to the default dictionary. This will update the internal
        dictionary. The dictionary should look like this::

            {
                'tcp': 'proto=tcp',
                'udp': 'proto=udp'
            }

    hidden_opts
        A list of mount options that will be ignored when considering a remount
        as part of the state application

        .. versionadded:: 2015.8.2

    bind_mount_copy_active_opts
        If set to ``False``, this option disables the default behavior of
        copying the options from the bind mount if it was found to be active.

        .. versionadded:: 3006.0
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    update_mount_cache = False

    if not name:
        ret["result"] = False
        ret["comment"] = "Must provide name to mount.mounted"
        return ret

    if not device:
        ret["result"] = False
        ret["comment"] = "Must provide device to mount.mounted"
        return ret

    if not fstype:
        ret["result"] = False
        ret["comment"] = "Must provide fstype to mount.mounted"
        return ret

    if device_name_regex is None:
        device_name_regex = []

    # Defaults is not a valid option on Mac OS
    if __grains__["os"] in ["MacOS", "Darwin"] and opts == "defaults":
        opts = "noowners"

    # Defaults is not a valid option on AIX
    if __grains__["os"] in ["AIX"]:
        if opts == "defaults":
            opts = ""

    # Make sure that opts is correct, it can be a list or a comma delimited
    # string
    if isinstance(opts, str):
        opts = opts.split(",")

    if isinstance(hidden_opts, str):
        hidden_opts = hidden_opts.split(",")

    # remove possible trailing slash
    if not name == "/":
        name = name.rstrip("/")

    device_list = []
    # Get the active data
    active = __salt__["mount.active"](extended=True)
    real_name = os.path.realpath(name)
    # real_name for comparisons to the active mount list
    comp_real_name = real_name.replace(" ", "\\040")
    if device.startswith("/"):
        if "bind" in opts and comp_real_name in active:
            _device = device.replace(" ", "\\040")
            if active[comp_real_name]["device"].startswith("/"):
                # Find the device that the bind really points at.
                while True:
                    if _device in active:
                        _real_device = active[_device]["device"]
                        if bind_mount_copy_active_opts:
                            opts = sorted(
                                set(
                                    opts
                                    + active[_device]["opts"]
                                    + active[_device]["superopts"]
                                )
                            )
                        active[comp_real_name]["opts"].append("bind")
                        break
                    _device = os.path.dirname(_device.replace("\\040", " "))
                real_device = _real_device
            else:
                # Remote file systems act differently.
                if _device in active:
                    if bind_mount_copy_active_opts:
                        opts = sorted(
                            set(
                                opts
                                + active[_device]["opts"]
                                + active[_device]["superopts"]
                            )
                        )
                    active[comp_real_name]["opts"].append("bind")
                real_device = active[comp_real_name]["device"]
        else:
            real_device = os.path.realpath(device)
    elif device.upper().startswith("UUID="):
        real_device = device.split("=")[1].strip('"').lower()
    elif device.upper().startswith("LABEL="):
        _label = device.split("=")[1]
        cmd = f"blkid -t LABEL={_label}"
        res = __salt__["cmd.run_all"](f"{cmd}")
        if res["retcode"] > 0:
            ret["comment"] = f"Unable to find device with label {_label}."
            ret["result"] = False
            return ret
        else:
            # output is a list of entries like this:
            # /dev/sda: LABEL="<label>" UUID="<uuid>" UUID_SUB="<uuid>" TYPE="btrfs"
            # exact list of properties varies between filesystems, but we're
            # only interested in the device in the first column
            for line in res["stdout"]:
                dev_with_label = line.split(":")[0]
                device_list.append(dev_with_label)
            real_device = device_list[0]
    else:
        real_device = device

    # LVS devices have 2 names under /dev:
    # /dev/mapper/vg--name-lv--name and /dev/vg-name/lv-name
    # No matter what name is used for mounting,
    # mount always displays the device as /dev/mapper/vg--name-lv--name
    # Note the double-dash escaping.
    # So, let's call that the canonical device name
    # We should normalize names of the /dev/vg-name/lv-name type to the canonical name
    lvs_match = re.match(r"^/dev/(?P<vg_name>[^/]+)/(?P<lv_name>[^/]+$)", device)
    if lvs_match:
        double_dash_escaped = {
            k: re.sub(r"-", "--", v) for k, v in lvs_match.groupdict().items()
        }
        mapper_device = "/dev/mapper/{vg_name}-{lv_name}".format(**double_dash_escaped)
        if os.path.exists(mapper_device):
            real_device = mapper_device

    # When included in a Salt state file, FUSE devices are prefaced by the
    # filesystem type and a hash, e.g. sshfs.  In the mount list only the
    # hostname is included.  So if we detect that the device is a FUSE device
    # then we remove the prefaced string so that the device in state matches
    # the device in the mount list.
    fuse_match = re.match(r"^\w+\#(?P<device_name>.+)", device)
    if fuse_match:
        if "device_name" in fuse_match.groupdict():
            real_device = fuse_match.group("device_name")

    if comp_real_name in active:
        if "superopts" not in active[comp_real_name]:
            active[comp_real_name]["superopts"] = []
        if mount:
            device_list.append(active[comp_real_name]["device"])
            device_list.append(os.path.realpath(device_list[0]))
            alt_device = (
                active[comp_real_name]["alt_device"]
                if "alt_device" in active[comp_real_name]
                else None
            )
            uuid_device = (
                active[comp_real_name]["device_uuid"]
                if "device_uuid" in active[comp_real_name]
                else None
            )
            label_device = (
                active[comp_real_name]["device_label"]
                if "device_label" in active[comp_real_name]
                else None
            )
            if alt_device and alt_device not in device_list:
                device_list.append(alt_device)
            if uuid_device and uuid_device not in device_list:
                device_list.append(uuid_device)
            if label_device and label_device not in device_list:
                device_list.append(label_device)
            if opts:
                mount_invisible_options = [
                    "_netdev",
                    "actimeo",
                    "bg",
                    "comment",
                    "defaults",
                    "delay_connect",
                    "direct-io-mode",
                    "intr",
                    "loop",
                    "nointr",
                    "nobootwait",
                    "nofail",
                    "password",
                    "reconnect",
                    "retry",
                    "soft",
                    "auto",
                    "users",
                    "bind",
                    "nonempty",
                    "transform_symlinks",
                    "port",
                    "backup-volfile-servers",
                ]

                if extra_mount_invisible_options:
                    mount_invisible_options.extend(extra_mount_invisible_options)

                if hidden_opts:
                    mount_invisible_options = list(
                        set(mount_invisible_options) | set(hidden_opts)
                    )

                # options which are provided as key=value (e.g. password=Zohp5ohb)
                mount_invisible_keys = [
                    "actimeo",
                    "comment",
                    "credentials",
                    "direct-io-mode",
                    "password",
                    "port",
                    "retry",
                    "secretfile",
                ]

                if extra_mount_invisible_keys:
                    mount_invisible_keys.extend(extra_mount_invisible_keys)

                # Some filesystems have options which should not force a remount.
                mount_ignore_fs_keys = {"ramfs": ["size"]}

                if extra_mount_ignore_fs_keys:
                    mount_ignore_fs_keys.update(extra_mount_ignore_fs_keys)

                # Some options are translated once mounted
                mount_translate_options = {
                    "tcp": "proto=tcp",
                    "udp": "proto=udp",
                }

                if extra_mount_translate_options:
                    mount_translate_options.update(extra_mount_translate_options)

                trigger_remount = []
                for opt in opts:
                    if opt in mount_translate_options:
                        opt = mount_translate_options[opt]

                    keyval_option = opt.split("=")[0]
                    if keyval_option in mount_invisible_keys:
                        opt = keyval_option

                    size_match = re.match(
                        r"size=(?P<size_value>[0-9]+)(?P<size_unit>k|m|g)", opt
                    )
                    if size_match:
                        converted_size = _size_convert(size_match)
                        opt = f"size={converted_size}k"
                    # make cifs option user synonym for option username which is reported by /proc/mounts
                    if fstype in ["cifs"] and opt.split("=")[0] == "user":
                        opt = "username={}".format(opt.split("=")[1])

                    if opt.split("=")[0] in mount_ignore_fs_keys.get(fstype, []):
                        opt = opt.split("=")[0]

                    # convert uid/gid to numeric value from user/group name
                    name_id_opts = {"uid": "user.info", "gid": "group.info"}
                    if opt.split("=")[0] in name_id_opts and len(opt.split("=")) > 1:
                        _givenid = opt.split("=")[1]
                        _param = opt.split("=")[0]
                        _id = _givenid
                        if not re.match("[0-9]+$", _givenid):
                            _info = __salt__[name_id_opts[_param]](_givenid)
                            if _info and _param in _info:
                                _id = _info[_param]
                        opt = _param + "=" + str(_id)

                    _active_superopts = active[comp_real_name].get("superopts", [])
                    for _active_opt in _active_superopts:
                        size_match = re.match(
                            r"size=(?P<size_value>[0-9]+)(?P<size_unit>k|m|g)",
                            _active_opt,
                        )
                        if size_match:
                            converted_size = _size_convert(size_match)
                            opt = f"size={converted_size}k"
                            _active_superopts.remove(_active_opt)
                            _active_opt = f"size={converted_size}k"
                            _active_superopts.append(_active_opt)

                    if (
                        opt not in active[comp_real_name]["opts"]
                        and opt not in _active_superopts
                        and opt not in mount_invisible_options
                        and opt not in mount_ignore_fs_keys.get(fstype, [])
                        and opt not in mount_invisible_keys
                    ):
                        trigger_remount.append(opt)

                if trigger_remount:
                    if __opts__["test"]:
                        ret["result"] = None
                        ret["comment"] = (
                            "Remount would be forced because options ({}) changed".format(
                                ",".join(sorted(trigger_remount))
                            )
                        )
                        return ret
                    else:
                        # Some file systems require umounting and mounting if options change
                        # add others to list that require similiar functionality
                        if fstype in ["nfs", "cvfs"] or fstype.startswith("fuse"):
                            ret["changes"]["umount"] = (
                                "Forced unmount and mount because "
                                + "options ({}) changed".format(
                                    ",".join(sorted(trigger_remount))
                                )
                            )
                            unmount_result = __salt__["mount.umount"](real_name)
                            if unmount_result is True:
                                mount_result = __salt__["mount.mount"](
                                    real_name,
                                    device,
                                    mkmnt=mkmnt,
                                    fstype=fstype,
                                    opts=opts,
                                )
                                ret["result"] = mount_result
                            else:
                                ret["result"] = False
                                ret["comment"] = "Unable to unmount {}: {}.".format(
                                    real_name, unmount_result
                                )
                                return ret
                        else:
                            ret["changes"]["umount"] = (
                                "Forced remount because "
                                + "options ({}) changed".format(
                                    ",".join(sorted(trigger_remount))
                                )
                            )
                            remount_result = __salt__["mount.remount"](
                                real_name,
                                device,
                                mkmnt=mkmnt,
                                fstype=fstype,
                                opts=opts,
                            )
                            ret["result"] = remount_result
                            # Cleanup after the remount, so we
                            # don't write remount into fstab
                            if "remount" in opts:
                                opts.remove("remount")

                        # Update the cache
                        update_mount_cache = True

                mount_cache = __salt__["mount.read_mount_cache"](real_name)
                if "opts" in mount_cache:
                    _missing = [opt for opt in mount_cache["opts"] if opt not in opts]

                    if _missing:
                        if __opts__["test"]:
                            ret["result"] = None
                            ret["comment"] = (
                                "Remount would be forced because"
                                " options ({})"
                                "changed".format(",".join(_missing))
                            )
                            return ret
                        else:
                            # Some file systems require umounting and mounting if options change
                            # add others to list that require similiar functionality
                            if fstype in ["nfs", "cvfs"] or fstype.startswith("fuse"):
                                ret["changes"]["umount"] = (
                                    "Forced unmount and mount because "
                                    + f"options ({opt}) changed"
                                )
                                unmount_result = __salt__["mount.umount"](real_name)
                                if unmount_result is True:
                                    mount_result = __salt__["mount.mount"](
                                        real_name,
                                        device,
                                        mkmnt=mkmnt,
                                        fstype=fstype,
                                        opts=opts,
                                    )
                                    ret["result"] = mount_result
                                else:
                                    ret["result"] = False
                                    ret["comment"] = "Unable to unmount {}: {}.".format(
                                        real_name, unmount_result
                                    )
                                    return ret
                            else:
                                ret["changes"]["umount"] = (
                                    "Forced remount because "
                                    + f"options ({opt}) changed"
                                )
                                remount_result = __salt__["mount.remount"](
                                    real_name,
                                    device,
                                    mkmnt=mkmnt,
                                    fstype=fstype,
                                    opts=opts,
                                )
                                ret["result"] = remount_result
                                # Cleanup after the remount, so we
                                # don't write remount into fstab
                                if "remount" in opts:
                                    opts.remove("remount")

                        update_mount_cache = True
                else:
                    update_mount_cache = True

            if real_device not in device_list:
                # name matches but device doesn't - need to umount
                _device_mismatch_is_ignored = None
                for regex in list(device_name_regex):
                    for _device in device_list:
                        if re.match(regex, _device):
                            _device_mismatch_is_ignored = _device
                            break
                if _device_mismatch_is_ignored:
                    ret["result"] = True
                    ret["comment"] = (
                        "An umount will not be forced "
                        + "because device matched device_name_regex: "
                        + _device_mismatch_is_ignored
                    )
                elif __opts__["test"]:
                    ret["result"] = None
                    ret["comment"] = (
                        "An umount would have been forced "
                        + "because devices do not match.  Watched: "
                        + device
                    )
                else:
                    ret["changes"]["umount"] = (
                        "Forced unmount because devices "
                        + "don't match. Wanted: "
                        + device
                    )
                    if real_device != device:
                        ret["changes"]["umount"] += " (" + real_device + ")"
                    ret["changes"]["umount"] += ", current: " + ", ".join(device_list)
                    out = __salt__["mount.umount"](real_name, user=user)
                    active = __salt__["mount.active"](extended=True)
                    if comp_real_name in active:
                        ret["comment"] = "Unable to unmount"
                        ret["result"] = None
                        return ret
                    update_mount_cache = True
            else:
                ret["comment"] = "Target was already mounted"
    # using a duplicate check so I can catch the results of a umount
    if comp_real_name not in active:
        if mount:
            # The mount is not present! Mount it
            if __opts__["test"]:
                ret["result"] = None
                if os.path.exists(name):
                    ret["comment"] = f"{name} would be mounted"
                elif mkmnt:
                    ret["comment"] = f"{name} would be created and mounted"
                else:
                    ret["comment"] = f"{name} does not exist and would not be created"
                return ret

            if not os.path.exists(name) and not mkmnt:
                ret["result"] = False
                ret["comment"] = "Mount directory is not present"
                return ret

            out = __salt__["mount.mount"](name, device, mkmnt, fstype, opts, user=user)
            active = __salt__["mount.active"](extended=True)
            update_mount_cache = True
            if isinstance(out, str):
                # Failed to (re)mount, the state has failed!
                ret["comment"] = out
                ret["result"] = False
                return ret
            elif comp_real_name in active:
                # (Re)mount worked!
                ret["comment"] = "Target was successfully mounted"
                ret["changes"]["mount"] = True
        elif not os.path.exists(name):
            if __opts__["test"]:
                ret["result"] = None
                if mkmnt:
                    ret["comment"] = f"{name} would be created, but not mounted"
                else:
                    ret["comment"] = (
                        "{} does not exist and would neither be created nor mounted".format(
                            name
                        )
                    )
            elif mkmnt:
                __salt__["file.mkdir"](name, user=user)
                ret["comment"] = f"{name} was created, not mounted"
            else:
                ret["comment"] = f"{name} not present and not mounted"
        else:
            if __opts__["test"]:
                ret["comment"] = f"{name} would not be mounted"
            else:
                ret["comment"] = f"{name} not mounted"

    if persist:
        if "/etc/fstab" == config:
            # Override default for Mac OS
            if __grains__["os"] in ["MacOS", "Darwin"]:
                config = "/etc/auto_salt"

            # Override default for AIX
            elif "AIX" in __grains__["os"]:
                config = "/etc/filesystems"

        if __opts__["test"]:
            if __grains__["os"] in ["MacOS", "Darwin"]:
                out = __salt__["mount.set_automaster"](
                    name, device, fstype, opts, config, test=True
                )
            elif __grains__["os"] in ["AIX"]:
                out = __salt__["mount.set_filesystems"](
                    name,
                    device,
                    fstype,
                    opts,
                    mount,
                    config,
                    test=True,
                    match_on=match_on,
                )
            else:
                out = __salt__["mount.set_fstab"](
                    name,
                    device,
                    fstype,
                    opts,
                    dump,
                    pass_num,
                    config,
                    test=True,
                    match_on=match_on,
                )
            if out != "present":
                ret["result"] = None
                if out == "new":
                    if mount:
                        comment = (
                            "{} is mounted, but needs to be "
                            "written to the fstab in order to be "
                            "made persistent.".format(name)
                        )
                    else:
                        comment = (
                            "{} needs to be "
                            "written to the fstab in order to be "
                            "made persistent.".format(name)
                        )
                elif out == "change":
                    if mount:
                        comment = "{} is mounted, but its fstab entry must be updated.".format(
                            name
                        )
                    else:
                        comment = f"The {name} fstab entry must be updated."
                else:
                    ret["result"] = False
                    comment = (
                        "Unable to detect fstab status for "
                        "mount point {} due to unexpected "
                        "output '{}' from call to "
                        "mount.set_fstab. This is most likely "
                        "a bug.".format(name, out)
                    )
                if "comment" in ret:
                    ret["comment"] = "{}. {}".format(ret["comment"], comment)
                else:
                    ret["comment"] = comment
                return ret

        else:
            if __grains__["os"] in ["MacOS", "Darwin"]:
                out = __salt__["mount.set_automaster"](
                    name, device, fstype, opts, config
                )
            elif __grains__["os"] in ["AIX"]:
                out = __salt__["mount.set_filesystems"](
                    name, device, fstype, opts, mount, config, match_on=match_on
                )
            else:
                out = __salt__["mount.set_fstab"](
                    name,
                    device,
                    fstype,
                    opts,
                    dump,
                    pass_num,
                    config,
                    match_on=match_on,
                )

        if update_mount_cache:
            cache_result = __salt__["mount.write_mount_cache"](
                real_name, device, mkmnt=mkmnt, fstype=fstype, mount_opts=opts
            )

        if out == "present":
            ret["comment"] += ". Entry already exists in the fstab."
            return ret
        if out == "new":
            ret["changes"]["persist"] = "new"
            ret["comment"] += ". Added new entry to the fstab."
            return ret
        if out == "change":
            ret["changes"]["persist"] = "update"
            ret["comment"] += ". Updated the entry in the fstab."
            return ret
        if out == "bad config":
            ret["result"] = False
            ret["comment"] += ". However, the fstab was not found."
            return ret

    return ret


def swap(name, persist=True, config="/etc/fstab"):
    """
    Activates a swap device

    .. code-block:: yaml

        /root/swapfile:
          mount.swap

    .. note::
        ``swap`` does not currently support LABEL
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    on_ = __salt__["mount.swaps"]()

    if __salt__["file.is_link"](name):
        real_swap_device = __salt__["file.readlink"](name)
        if not real_swap_device.startswith("/"):
            real_swap_device = f"/dev/{os.path.basename(real_swap_device)}"
    else:
        real_swap_device = name

    if real_swap_device in on_:
        ret["comment"] = f"Swap {name} already active"
    elif __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Swap {name} is set to be activated"
    else:
        __salt__["mount.swapon"](real_swap_device)

        on_ = __salt__["mount.swaps"]()

        if real_swap_device in on_:
            ret["comment"] = f"Swap {name} activated"
            ret["changes"] = on_[real_swap_device]
        else:
            ret["comment"] = f"Swap {name} failed to activate"
            ret["result"] = False

    if persist:
        device_key_name = "device"
        if "AIX" in __grains__["os"]:
            device_key_name = "dev"
            if "/etc/fstab" == config:
                # Override default for AIX
                config = "/etc/filesystems"
            fstab_data = __salt__["mount.filesystems"](config)
        else:
            fstab_data = __salt__["mount.fstab"](config)
        if __opts__["test"]:
            if name not in fstab_data and name not in [
                fstab_data[item]["device"] for item in fstab_data
            ]:
                ret["result"] = None
                if name in on_:
                    ret["comment"] = (
                        "Swap {} is set to be added to the fstab and to be activated".format(
                            name
                        )
                    )
            return ret

        if "none" in fstab_data:
            if (
                fstab_data["none"][device_key_name] == name
                and fstab_data["none"]["fstype"] != "swap"
            ):
                return ret

        if "AIX" in __grains__["os"]:
            out = None
            ret["result"] = False
            ret["comment"] += ". swap not present in /etc/filesystems on AIX."
            return ret
        else:
            # present, new, change, bad config
            # Make sure the entry is in the fstab
            out = __salt__["mount.set_fstab"](
                "none", name, "swap", ["defaults"], 0, 0, config
            )
        if out == "present":
            return ret
        if out == "new":
            ret["changes"]["persist"] = "new"
            ret["comment"] += ". Added new entry to the fstab."
            return ret
        if out == "change":
            ret["changes"]["persist"] = "update"
            ret["comment"] += ". Updated the entry in the fstab."
            return ret
        if out == "bad config":
            ret["result"] = False
            ret["comment"] += ". However, the fstab was not found."
            return ret
    return ret


def unmounted(
    name, device=None, config="/etc/fstab", persist=False, user=None, **kwargs
):
    """
    .. versionadded:: 0.17.0

    Verify that a device is not mounted

    name
        The path to the location where the device is to be unmounted from

    device
        The device to be unmounted.  This is optional because the device could
        be mounted in multiple places.

        .. versionadded:: 2015.5.0

    config
        Set an alternative location for the fstab, Default is ``/etc/fstab``

    persist
        Set if the mount should be purged from the fstab, Default is ``False``

    user
        The user to own the mount; this defaults to the user salt is
        running as on the minion
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    update_mount_cache = False

    if not name:
        ret["result"] = False
        ret["comment"] = "Must provide name to mount.unmounted"
        return ret

    # Get the active data
    active = __salt__["mount.active"](extended=True)
    comp_name = name.replace(" ", "\\040")
    if comp_name not in active:
        # Nothing to unmount
        ret["comment"] = "Target was already unmounted"
    if comp_name in active:
        # The mount is present! Unmount it
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Mount point {name} is mounted but should not be"
            return ret
        if device:
            out = __salt__["mount.umount"](name, device, user=user)
            update_mount_cache = True
        else:
            out = __salt__["mount.umount"](name, user=user)
            update_mount_cache = True
        if isinstance(out, str):
            # Failed to umount, the state has failed!
            ret["comment"] = out
            ret["result"] = False
        elif out is True:
            # umount worked!
            ret["comment"] = "Target was successfully unmounted"
            ret["changes"]["umount"] = True
        else:
            ret["comment"] = "Execute set to False, Target was not unmounted"
            ret["result"] = True

    if update_mount_cache:
        cache_result = __salt__["mount.delete_mount_cache"](name)

    if persist:
        device_key_name = "device"
        # Override default for Mac OS
        if __grains__["os"] in ["MacOS", "Darwin"] and config == "/etc/fstab":
            config = "/etc/auto_salt"
            fstab_data = __salt__["mount.automaster"](config)
        elif "AIX" in __grains__["os"]:
            device_key_name = "dev"
            if config == "/etc/fstab":
                config = "/etc/filesystems"
            fstab_data = __salt__["mount.filesystems"](config)
        else:
            fstab_data = __salt__["mount.fstab"](config)

        if name not in fstab_data:
            ret["comment"] += ". fstab entry not found"
        else:
            if device:
                if fstab_data[name][device_key_name] != device:
                    ret["comment"] += ". fstab entry for device {} not found".format(
                        device
                    )
                    return ret
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = (
                    "Mount point {} is unmounted but needs to "
                    "be purged from {} to be made "
                    "persistent".format(name, config)
                )
                return ret
            else:
                if __grains__["os"] in ["MacOS", "Darwin"]:
                    out = __salt__["mount.rm_automaster"](name, device, config)
                elif "AIX" in __grains__["os"]:
                    out = __salt__["mount.rm_filesystems"](name, device, config)
                else:
                    out = __salt__["mount.rm_fstab"](name, device, config)
                if out is not True:
                    ret["result"] = False
                    ret["comment"] += ". Failed to persist purge"
                else:
                    ret["comment"] += ". Removed target from fstab"
                    ret["changes"]["persist"] = "purged"

    return ret


def mod_watch(name, user=None, **kwargs):
    """
    The mounted watcher, called to invoke the watch command.

    .. note::
        This state exists to support special handling of the ``watch``
        :ref:`requisite <requisites>`. It should not be called directly.

        Parameters for this function should be set by the state being triggered.

    name
        The name of the mount point

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if kwargs["sfun"] == "mounted":
        out = __salt__["mount.remount"](
            name, kwargs["device"], False, kwargs["fstype"], kwargs["opts"], user=user
        )
        if out:
            ret["comment"] = f"{name} remounted"
        else:
            ret["result"] = False
            ret["comment"] = f"{name} failed to remount: {out}"
    else:
        ret["comment"] = "Watch not supported in {} at this time".format(kwargs["sfun"])
    return ret


def _convert_to(maybe_device, convert_to):
    """
    Convert a device name, UUID or LABEL to a device name, UUID or
    LABEL.

    Return the fs_spec required for fstab.

    """

    # Fast path. If we already have the information required, we can
    # save one blkid call
    if (
        not convert_to
        or (convert_to == "device" and maybe_device.startswith("/"))
        or maybe_device.startswith(f"{convert_to.upper()}=")
    ):
        return maybe_device

    # Get the device information
    if maybe_device.startswith("/"):
        blkid = __salt__["disk.blkid"](maybe_device)
    else:
        blkid = __salt__["disk.blkid"](token=maybe_device)

    result = None
    if len(blkid) == 1:
        if convert_to == "device":
            result = next(iter(blkid))
        else:
            key = convert_to.upper()
            result = f"{key}={next(iter(blkid.values()))[key]}"

    return result


def fstab_present(
    name,
    fs_file,
    fs_vfstype,
    fs_mntops="defaults",
    fs_freq=0,
    fs_passno=0,
    mount_by=None,
    config="/etc/fstab",
    mount=True,
    match_on="auto",
    not_change=False,
    fs_mount=True,
):
    """Makes sure that a fstab mount point is present.

    name
        The name of block device. Can be any valid fs_spec value.

    fs_file
        Mount point (target) for the filesystem.

    fs_vfstype
        The type of the filesystem (e.g. ext4, xfs, btrfs, ...)

    fs_mntops
        The mount options associated with the filesystem. Default is
        ``defaults``.

    fs_freq
        Field is used by dump to determine which fs need to be
        dumped. Default is ``0``

    fs_passno
        Field is used by fsck to determine the order in which
        filesystem checks are done at boot time. Default is ``0``

    fs_mount
        Field is used only in AIX systems to determine if the
        filesystem will be mounted by ``mount all``

    mount_by
        Select the final value for fs_spec. Can be [``None``,
        ``device``, ``label``, ``uuid``, ``partlabel``,
        ``partuuid``]. If ``None``, the value for fs_spect will be the
        parameter ``name``, in other case will search the correct
        value based on the device name. For example, for ``uuid``, the
        value for fs_spec will be of type 'UUID=xxx' instead of the
        device name set in ``name``.

    config
        Place where the fstab file lives. Default is ``/etc/fstab``

    mount
        Set if the mount should be mounted immediately. Default is
        ``True``

    match_on
        A name or list of fstab properties on which this state should
        be applied.  Default is ``auto``, a special value indicating
        to guess based on fstype.  In general, ``auto`` matches on
        name for recognized special devices and device otherwise.

    not_change
        By default, if the entry is found in the fstab file but is
        different from the expected content (like different options),
        the entry will be replaced with the correct content. If this
        parameter is set to ``True`` and the line is found, the
        original content will be preserved.

    """
    ret = {
        "name": name,
        "result": False,
        "changes": {},
        "comment": [],
    }

    # Adjust fs_mntops based on the OS
    if fs_mntops == "defaults":
        if __grains__["os"] in ["MacOS", "Darwin"]:
            fs_mntops = "noowners"
        elif __grains__["os"] == "AIX":
            fs_mntops = ""

    # Adjust the config file based on the OS
    if config == "/etc/fstab":
        if __grains__["os"] in ["MacOS", "Darwin"]:
            config = "/etc/auto_salt"
        elif __grains__["os"] == "AIX":
            config = "/etc/filesystems"

    if not fs_file == "/":
        fs_file = fs_file.rstrip("/")

    fs_spec = _convert_to(name, mount_by)

    # Validate that the device is valid after the conversion
    if not fs_spec:
        msg = "Device {} cannot be converted to {}"
        ret["comment"].append(msg.format(name, mount_by))
        return ret

    if __opts__["test"]:
        if __grains__["os"] in ["MacOS", "Darwin"]:
            out = __salt__["mount.set_automaster"](
                name=fs_file,
                device=fs_spec,
                fstype=fs_vfstype,
                opts=fs_mntops,
                config=config,
                test=True,
                not_change=not_change,
            )
        elif __grains__["os"] == "AIX":
            out = __salt__["mount.set_filesystems"](
                name=fs_file,
                device=fs_spec,
                fstype=fs_vfstype,
                opts=fs_mntops,
                mount=fs_mount,
                config=config,
                test=True,
                match_on=match_on,
                not_change=not_change,
            )
        else:
            out = __salt__["mount.set_fstab"](
                name=fs_file,
                device=fs_spec,
                fstype=fs_vfstype,
                opts=fs_mntops,
                dump=fs_freq,
                pass_num=fs_passno,
                config=config,
                test=True,
                match_on=match_on,
                not_change=not_change,
            )
        ret["result"] = None
        if out == "present":
            msg = "{} entry is already in {}."
            ret["comment"].append(msg.format(fs_file, config))
        elif out == "new":
            msg = "{} entry will be written in {}."
            ret["comment"].append(msg.format(fs_file, config))
            if mount:
                msg = f"Will mount {name} on {fs_file}"
                ret["comment"].append(msg)
        elif out == "change":
            msg = "{} entry will be updated in {}."
            ret["comment"].append(msg.format(fs_file, config))
        else:
            ret["result"] = False
            msg = "{} entry cannot be created in {}: {}."
            ret["comment"].append(msg.format(fs_file, config, out))
        return ret

    if __grains__["os"] in ["MacOS", "Darwin"]:
        out = __salt__["mount.set_automaster"](
            name=fs_file,
            device=fs_spec,
            fstype=fs_vfstype,
            opts=fs_mntops,
            config=config,
            not_change=not_change,
        )
    elif __grains__["os"] == "AIX":
        out = __salt__["mount.set_filesystems"](
            name=fs_file,
            device=fs_spec,
            fstype=fs_vfstype,
            opts=fs_mntops,
            mount=fs_mount,
            config=config,
            match_on=match_on,
            not_change=not_change,
        )
    else:
        out = __salt__["mount.set_fstab"](
            name=fs_file,
            device=fs_spec,
            fstype=fs_vfstype,
            opts=fs_mntops,
            dump=fs_freq,
            pass_num=fs_passno,
            config=config,
            match_on=match_on,
            not_change=not_change,
        )

    ret["result"] = True
    if out == "present":
        msg = "{} entry was already in {}."
        ret["comment"].append(msg.format(fs_file, config))
    elif out == "new":
        ret["changes"]["persist"] = out
        msg = "{} entry added in {}."
        ret["comment"].append(msg.format(fs_file, config))
        if mount:
            out = __salt__["mount.mount"](fs_file)
            if type(out) == str:
                ret["result"] = False
                msg = "Error while mounting {}".format(out.split(":", maxsplit=1)[1])
            else:
                msg = f"Mounted {name} on {fs_file}"
            ret["comment"].append(msg)
    elif out == "change":
        ret["changes"]["persist"] = out
        msg = "{} entry updated in {}."
        ret["comment"].append(msg.format(fs_file, config))
    else:
        ret["result"] = False
        msg = "{} entry cannot be changed in {}: {}."
        ret["comment"].append(msg.format(fs_file, config, out))

    return ret


def fstab_absent(name, fs_file, mount_by=None, config="/etc/fstab"):
    """
    Makes sure that a fstab mount point is absent.

    name
        The name of block device. Can be any valid fs_spec value.

    fs_file
        Mount point (target) for the filesystem.

    mount_by
        Select the final value for fs_spec. Can be [``None``,
        ``device``, ``label``, ``uuid``, ``partlabel``,
        ``partuuid``]. If ``None``, the value for fs_spect will be the
        parameter ``name``, in other case will search the correct
        value based on the device name. For example, for ``uuid``, the
        value for fs_spec will be of type 'UUID=xxx' instead of the
        device name set in ``name``.

    config
        Place where the fstab file lives

    """
    ret = {
        "name": name,
        "result": False,
        "changes": {},
        "comment": [],
    }

    # Adjust the config file based on the OS
    if config == "/etc/fstab":
        if __grains__["os"] in ["MacOS", "Darwin"]:
            config = "/etc/auto_salt"
        elif __grains__["os"] == "AIX":
            config = "/etc/filesystems"

    if not fs_file == "/":
        fs_file = fs_file.rstrip("/")

    fs_spec = _convert_to(name, mount_by)

    if __grains__["os"] in ["MacOS", "Darwin"]:
        fstab_data = __salt__["mount.automaster"](config)
    elif __grains__["os"] == "AIX":
        fstab_data = __salt__["mount.filesystems"](config)
    else:
        fstab_data = __salt__["mount.fstab"](config)

    if __opts__["test"]:
        ret["result"] = None
        if fs_file not in fstab_data:
            msg = "{} entry is already missing in {}."
            ret["comment"].append(msg.format(fs_file, config))
        else:
            msg = "{} entry will be removed from {}."
            ret["comment"].append(msg.format(fs_file, config))
        return ret

    if fs_file in fstab_data:
        if __grains__["os"] in ["MacOS", "Darwin"]:
            out = __salt__["mount.rm_automaster"](
                name=fs_file, device=fs_spec, config=config
            )
        elif __grains__["os"] == "AIX":
            out = __salt__["mount.rm_filesystems"](
                name=fs_file, device=fs_spec, config=config
            )
        else:
            out = __salt__["mount.rm_fstab"](
                name=fs_file, device=fs_spec, config=config
            )

        if out is not True:
            ret["result"] = False
            msg = "{} entry failed when removing from {}."
            ret["comment"].append(msg.format(fs_file, config))
        else:
            ret["result"] = True
            ret["changes"]["persist"] = "removed"
            msg = "{} entry removed from {}."
            ret["comment"].append(msg.format(fs_file, config))
    else:
        ret["result"] = True
        msg = "{} entry is already missing in {}."
        ret["comment"].append(msg.format(fs_file, config))

    return ret



"""
    Test cases for salt.modules.rvm
"""

import pytest

import salt.modules.rvm as rvm
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {
        rvm: {
            "__salt__": {
                "cmd.has_exec": MagicMock(return_value=True),
                "config.option": MagicMock(return_value=None),
            }
        }
    }


def test_rvm():
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(rvm.__salt__, {"cmd.run_all": mock}):
        rvm._rvm(["install", "1.9.3"])
        mock.assert_called_once_with(
            ["/usr/local/rvm/bin/rvm", "install", "1.9.3"],
            runas=None,
            cwd=None,
            python_shell=False,
            env=None,
        )


def test_rvm_do():
    mock = MagicMock(return_value={"retcode": 0, "stdout": "stdout"})
    with patch.dict(rvm.__salt__, {"cmd.run_all": mock}):
        rvm._rvm_do("1.9.3", ["gemset", "list"])
        mock.assert_called_once_with(
            ["/usr/local/rvm/bin/rvm", "1.9.3", "do", "gemset", "list"],
            runas=None,
            cwd=None,
            python_shell=False,
            env=None,
        )


def test_install():
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(rvm.__salt__, {"cmd.run_all": mock}):
        rvm.install()
        curl_cmd = (
            "curl -Ls"
            " https://raw.githubusercontent.com/rvm/rvm/master/binscripts/rvm-installer"
            " | bash -s stable"
        )
        mock.assert_called_once_with(curl_cmd, runas=None, python_shell=True)


def test_install_ruby_nonroot():
    mock = MagicMock(return_value={"retcode": 0, "stdout": "stdout"})
    expected = [
        call(
            ["/usr/local/rvm/bin/rvm", "autolibs", "disable", "2.0.0"],
            runas="rvm",
            cwd=None,
            python_shell=False,
            env=None,
        ),
        call(
            ["/usr/local/rvm/bin/rvm", "install", "2.0.0", "--disable-binary"],
            runas="rvm",
            cwd=None,
            python_shell=False,
            env=None,
        ),
    ]
    with patch.dict(rvm.__salt__, {"cmd.run_all": mock}):
        rvm.install_ruby("2.0.0", runas="rvm")
        assert mock.call_args_list == expected


def test_install_with_env():
    mock = MagicMock(return_value={"retcode": 0, "stdout": "stdout"})
    expected = [
        call(
            ["/usr/local/rvm/bin/rvm", "install", "2.0.0"],
            runas=None,
            cwd=None,
            python_shell=False,
            env=[{"RUBY_CONFIGURE_OPTS": "--foobar"}],
        )
    ]
    with patch.dict(rvm.__salt__, {"cmd.run_all": mock}):
        rvm.install_ruby("2.0.0", env=[{"RUBY_CONFIGURE_OPTS": "--foobar"}])
        assert mock.call_args_list == expected


def test_install_with_opts():
    mock = MagicMock(return_value={"retcode": 0, "stdout": "stdout"})
    expected = [
        call(
            [
                "/usr/local/rvm/bin/rvm",
                "install",
                "2.0.0",
                "-C --enable-shared,--with-readline-dir=$HOME/.rvm/usr",
                "--patch /path/to/awesome.patch",
            ],
            runas=None,
            cwd=None,
            python_shell=False,
            env=None,
        )
    ]
    with patch.dict(rvm.__salt__, {"cmd.run_all": mock}):
        rvm.install_ruby(
            "2.0.0",
            opts=[
                "-C --enable-shared,--with-readline-dir=$HOME/.rvm/usr",
                "--patch /path/to/awesome.patch",
            ],
        )
        assert mock.call_args_list == expected


def test_list():
    list_output = """
rvm rubies

   jruby-1.6.5.1 [ amd64 ]
   ree-1.8.7-2011.03 [ x86_64 ]
   ree-1.8.7-2011.12 [ x86_64 ]
=* ree-1.8.7-2012.02 [ x86_64 ]
   ruby-1.9.2-p180 [ x86_64 ]
   ruby-1.9.3-p125 [ x86_64 ]
   ruby-head [ x86_64 ]

# => - current
# =* - current && default
#  * - default

"""
    with patch.object(rvm, "_rvm") as mock_method:
        mock_method.return_value = list_output
        assert [
            ["jruby", "1.6.5.1", False],
            ["ree", "1.8.7-2011.03", False],
            ["ree", "1.8.7-2011.12", False],
            ["ree", "1.8.7-2012.02", True],
            ["ruby", "1.9.2-p180", False],
            ["ruby", "1.9.3-p125", False],
            ["ruby", "head", False],
        ] == rvm.list_()


def test_gemset_list():
    output = """
gemsets for ree-1.8.7-2012.02 (found in /usr/local/rvm/gems/ree-1.8.7-2012.02)
   global
   bar
   foo

"""
    with patch.object(rvm, "_rvm_do") as mock_method:
        mock_method.return_value = output
        assert ["global", "bar", "foo"] == rvm.gemset_list()


def test_gemset_list_all():
    output = """

gemsets for ruby-1.9.3-p125 (found in /usr/local/rvm/gems/ruby-1.9.3-p125)
   9bar
   9foo
   global


gemsets for ruby-head (found in /usr/local/rvm/gems/ruby-head)
   global
   headbar
   headfoo


gemsets for jruby-1.6.5.1 (found in /usr/local/rvm/gems/jruby-1.6.5.1)
   global
   jbar
   jfoo


gemsets for ruby-1.9.2-p180 (found in /usr/local/rvm/gems/ruby-1.9.2-p180)
   global


"""
    with patch.object(rvm, "_rvm_do") as mock_method:
        mock_method.return_value = output
        assert {
            "jruby-1.6.5.1": ["global", "jbar", "jfoo"],
            "ruby-1.9.2-p180": ["global"],
            "ruby-1.9.3-p125": ["9bar", "9foo", "global"],
            "ruby-head": ["global", "headbar", "headfoo"],
        } == rvm.gemset_list_all()



"""
Tests for win_file execution module
"""

import pytest
from saltfactories.utils import random_string

import salt.modules.cmdmod as cmd
import salt.modules.win_file as win_file
import salt.modules.win_useradd as win_useradd
import salt.utils.win_dacl as win_dacl
from salt.exceptions import CommandExecutionError
from tests.support.mock import patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {
        win_file: {
            "__utils__": {
                "dacl.check_perms": win_dacl.check_perms,
                "dacl.set_perms": win_dacl.set_perms,
            },
            "__opts__": {"test": False},
        },
        win_useradd: {
            "__salt__": {
                "cmd.run_all": cmd.run_all,
            },
        },
    }


@pytest.fixture
def temp_account():
    user_name = random_string("test-account-", uppercase=False)
    with pytest.helpers.create_account(username=user_name) as account:
        win_useradd.addgroup(account.username, "Users")
        yield account.username


@pytest.fixture(scope="function")
def test_file():
    with pytest.helpers.temp_file("win_file_test.file") as test_file:
        yield test_file


def test_check_perms_set_owner_test_true(test_file):
    """
    Test setting the owner of a file with test=True
    """
    expected = {
        "comment": "",
        "changes": {"owner": "Backup Operators"},
        "name": str(test_file),
        "result": None,
    }
    with patch.dict(win_file.__opts__, {"test": True}):
        result = win_file.check_perms(
            path=str(test_file), owner="Backup Operators", inheritance=None
        )
        assert result == expected


def test_check_perms_set_owner(test_file):
    """
    Test setting the owner of a file
    """
    expected = {
        "comment": "",
        "changes": {"owner": "Backup Operators"},
        "name": str(test_file),
        "result": True,
    }
    result = win_file.check_perms(
        path=str(test_file), owner="Backup Operators", inheritance=None
    )
    assert result == expected


def test_check_perms_deny_test_true(test_file):
    """
    Test setting deny perms on a file with test=True
    """
    expected = {
        "comment": "",
        "changes": {"deny_perms": {"Users": {"permissions": "read_execute"}}},
        "name": str(test_file),
        "result": None,
    }
    with patch.dict(win_file.__opts__, {"test": True}):
        result = win_file.check_perms(
            path=str(test_file),
            deny_perms={"Users": {"perms": "read_execute"}},
            inheritance=None,
        )
    assert result == expected


def test_check_perms_deny(test_file):
    """
    Test setting deny perms on a file
    """
    expected = {
        "comment": "",
        "changes": {"deny_perms": {"Users": {"permissions": "read_execute"}}},
        "name": str(test_file),
        "result": True,
    }
    result = win_file.check_perms(
        path=str(test_file),
        deny_perms={"Users": {"perms": "read_execute"}},
        inheritance=None,
    )
    assert result == expected


def test_check_perms_grant_test_true(test_file):
    """
    Test setting grant perms on a file with test=True
    """
    expected = {
        "comment": "",
        "changes": {"grant_perms": {"Users": {"permissions": "read_execute"}}},
        "name": str(test_file),
        "result": None,
    }
    with patch.dict(win_file.__opts__, {"test": True}):
        result = win_file.check_perms(
            path=str(test_file),
            grant_perms={"Users": {"perms": "read_execute"}},
            inheritance=None,
        )
        assert result == expected


def test_check_perms_grant(test_file):
    """
    Test setting grant perms on a file
    """
    expected = {
        "comment": "",
        "changes": {"grant_perms": {"Users": {"permissions": "read_execute"}}},
        "name": str(test_file),
        "result": True,
    }
    result = win_file.check_perms(
        path=str(test_file),
        grant_perms={"Users": {"perms": "read_execute"}},
        inheritance=None,
    )
    assert result == expected


def test_check_perms_inheritance_false_test_true(test_file):
    """
    Test setting inheritance to False with test=True
    """
    expected = {
        "comment": "",
        "changes": {"inheritance": False},
        "name": str(test_file),
        "result": None,
    }
    with patch.dict(win_file.__opts__, {"test": True}):
        result = win_file.check_perms(path=str(test_file), inheritance=False)
    assert result == expected


def test_check_perms_inheritance_false(test_file):
    """
    Test setting inheritance to False
    """
    expected = {
        "comment": "",
        "changes": {"inheritance": False},
        "name": str(test_file),
        "result": True,
    }
    result = win_file.check_perms(path=str(test_file), inheritance=False)
    assert result == expected


def test_check_perms_inheritance_true(test_file):
    """
    Test setting inheritance to true when it's already true (default)
    """
    expected = {
        "comment": "",
        "changes": {},
        "name": str(test_file),
        "result": True,
    }
    result = win_file.check_perms(path=str(test_file), inheritance=True)
    assert result == expected


def test_check_perms_reset_test_true(test_file, temp_account):
    """
    Test resetting perms with test=True. This shows minimal changes
    """
    # Turn off inheritance
    win_dacl.set_inheritance(obj_name=str(test_file), enabled=False, clear=True)
    # Set some permissions
    win_dacl.set_permissions(
        obj_name=str(test_file),
        principal=temp_account,
        permissions="full_control",
    )
    expected = {
        "comment": "",
        "changes": {
            "grant_perms": {
                "Administrators": {"permissions": "full_control"},
                "Users": {"permissions": "read_execute"},
            },
            "remove_perms": {
                f"{temp_account}": {
                    "grant": {
                        "applies to": "This folder only",
                        "permissions": "Full control",
                    }
                }
            },
        },
        "name": str(test_file),
        "result": None,
    }
    with patch.dict(win_file.__opts__, {"test": True}):
        result = win_file.check_perms(
            path=str(test_file),
            grant_perms={
                "Users": {"perms": "read_execute"},
                "Administrators": {"perms": "full_control"},
            },
            inheritance=False,
            reset=True,
        )
        assert result == expected


def test_check_perms_reset(test_file, temp_account):
    """
    Test resetting perms on a File
    """
    # Turn off inheritance
    win_dacl.set_inheritance(obj_name=str(test_file), enabled=False, clear=True)
    # Set some permissions
    win_dacl.set_permissions(
        obj_name=str(test_file),
        principal=temp_account,
        permissions="full_control",
    )
    expected = {
        "comment": "",
        "changes": {
            "grant_perms": {
                "Administrators": {"permissions": "full_control"},
                "Users": {"permissions": "read_execute"},
            },
            "remove_perms": {
                f"{temp_account}": {
                    "grant": {
                        "applies to": "This folder only",
                        "permissions": "Full control",
                    }
                }
            },
        },
        "name": str(test_file),
        "result": True,
    }
    result = win_file.check_perms(
        path=str(test_file),
        grant_perms={
            "Users": {"perms": "read_execute"},
            "Administrators": {"perms": "full_control"},
        },
        inheritance=False,
        reset=True,
    )
    assert result == expected


def test_check_perms_issue_43328(test_file):
    """
    Make sure that a CommandExecutionError is raised if the file does NOT
    exist
    """
    fake_file = test_file.parent / "fake.file"
    with pytest.raises(CommandExecutionError):
        win_file.check_perms(str(fake_file))



from textwrap import dedent

import pytest

import salt.modules.ethtool as ethtool
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        ethtool: {
            "__salt__": {},
        }
    }


@pytest.fixture(scope="module")
def pause_ret():
    cmdret = dedent(
        """Pause parameters for eth0:
        Autonegotiate:  on
        RX:             on
        TX:             on
        RX negotiated:  off
        TX negotiated:  off"""
    )
    return cmdret


@pytest.fixture(scope="module")
def features_ret():
    cmdret = dedent(
        """Features for eth0:
        rx-checksumming: on [fixed]
        tx-checksumming: on
                tx-checksum-ipv4: off [fixed]
                tx-checksum-ip-generic: on
                tx-checksum-ipv6: off [fixed]
                tx-checksum-fcoe-crc: off [fixed]
                tx-checksum-sctp: off [fixed]
        scatter-gather: on
                tx-scatter-gather: on
                tx-scatter-gather-fraglist: off [fixed]
        tcp-segmentation-offload: on
                tx-tcp-segmentation: on
                tx-tcp-ecn-segmentation: on
                tx-tcp-mangleid-segmentation: off
                tx-tcp6-segmentation: on
        udp-fragmentation-offload: off
        generic-segmentation-offload: on
        generic-receive-offload: on
        large-receive-offload: off [fixed]
        rx-vlan-offload: off [fixed]
        tx-vlan-offload: off [fixed]
        ntuple-filters: off [fixed]
        receive-hashing: off [fixed]
        highdma: on [fixed]
        rx-vlan-filter: on [fixed]
        vlan-challenged: off [fixed]
        tx-lockless: off [fixed]
        netns-local: off [fixed]
        tx-gso-robust: on [fixed]
        tx-fcoe-segmentation: off [fixed]
        tx-gre-segmentation: off [fixed]
        tx-gre-csum-segmentation: off [fixed]
        tx-ipxip4-segmentation: off [fixed]
        tx-ipxip6-segmentation: off [fixed]
        tx-udp_tnl-segmentation: off [fixed]
        tx-udp_tnl-csum-segmentation: off [fixed]
        tx-gso-partial: off [fixed]
        tx-sctp-segmentation: off [fixed]
        tx-esp-segmentation: off [fixed]
        tx-udp-segmentation: off [fixed]
        fcoe-mtu: off [fixed]
        tx-nocache-copy: off
        loopback: off [fixed]
        rx-fcs: off [fixed]
        rx-all: off [fixed]
        tx-vlan-stag-hw-insert: off [fixed]
        rx-vlan-stag-hw-parse: off [fixed]
        rx-vlan-stag-filter: off [fixed]
        l2-fwd-offload: off [fixed]
        hw-tc-offload: off [fixed]
        esp-hw-offload: off [fixed]
        esp-tx-csum-hw-offload: off [fixed]
        rx-udp_tunnel-port-offload: off [fixed]
        tls-hw-tx-offload: off [fixed]
        tls-hw-rx-offload: off [fixed]
        rx-gro-hw: off [fixed]
        tls-hw-record: off [fixed]"""
    )
    return cmdret


def test_ethtool__ethtool_command_which_fail():
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        with pytest.raises(CommandExecutionError):
            ethtool._ethtool_command("eth0")


def test_ethtool__ethtool_command_operation_not_supported():
    mock_cmd_run = MagicMock(
        side_effect=[
            "Pause parameters for eth0:\nCannot get device pause settings: Operation not supported",
            "Cannot get device pause settings: Operation not supported",
        ]
    )
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/ethtool")
    ), patch.dict(ethtool.__salt__, {"cmd.run": mock_cmd_run}):
        with pytest.raises(CommandExecutionError):
            ethtool._ethtool_command("eth0", "-a")
            ethtool._ethtool_command("eth0", "-A", autoneg="off", rx="off", tx="off")


def test_ethtool__ethtool_command(pause_ret):
    mock_cmd_run = MagicMock(return_value=pause_ret)

    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/ethtool")
    ), patch.dict(ethtool.__salt__, {"cmd.run": mock_cmd_run}):
        ret = ethtool._ethtool_command("eth0", "-A", autoneg="off", rx="off", tx="off")

    mock_cmd_run.assert_called_once_with(
        "/sbin/ethtool -A eth0 autoneg off rx off tx off", ignore_retcode=True
    )
    assert pause_ret.splitlines() == ret


def test_ethtool__validate_params():
    with pytest.raises(CommandExecutionError):
        ethtool._validate_params(["not_found"], {"eth": "tool"})
    assert ethtool._validate_params(["eth"], {"eth": "tool"}) == {"eth": "tool"}
    assert ethtool._validate_params(["eth", "not_found"], {"eth": "tool"}) == {
        "eth": "tool"
    }
    assert ethtool._validate_params(["eth", "salt"], {"eth": True, "salt": False}) == {
        "eth": "on",
        "salt": "off",
    }


def test_ethtool_show_pause(pause_ret):
    expected = {
        "Autonegotiate": True,
        "RX": True,
        "RX negotiated": False,
        "TX": True,
        "TX negotiated": False,
    }

    with patch(
        "salt.modules.ethtool._ethtool_command",
        MagicMock(return_value=pause_ret.splitlines()),
    ):
        ret = ethtool.show_pause("eth0")

    assert expected == ret


def test_ethtool_show_features(features_ret):
    expected = {
        "esp-hw-offload": {"fixed": True, "on": False},
        "esp-tx-csum-hw-offload": {"fixed": True, "on": False},
        "fcoe-mtu": {"fixed": True, "on": False},
        "generic-receive-offload": {"fixed": False, "on": True},
        "generic-segmentation-offload": {"fixed": False, "on": True},
        "highdma": {"fixed": True, "on": True},
        "hw-tc-offload": {"fixed": True, "on": False},
        "l2-fwd-offload": {"fixed": True, "on": False},
        "large-receive-offload": {"fixed": True, "on": False},
        "loopback": {"fixed": True, "on": False},
        "netns-local": {"fixed": True, "on": False},
        "ntuple-filters": {"fixed": True, "on": False},
        "receive-hashing": {"fixed": True, "on": False},
        "rx-all": {"fixed": True, "on": False},
        "rx-checksumming": {"fixed": True, "on": True},
        "rx-fcs": {"fixed": True, "on": False},
        "rx-gro-hw": {"fixed": True, "on": False},
        "rx-udp_tunnel-port-offload": {"fixed": True, "on": False},
        "rx-vlan-filter": {"fixed": True, "on": True},
        "rx-vlan-offload": {"fixed": True, "on": False},
        "rx-vlan-stag-filter": {"fixed": True, "on": False},
        "rx-vlan-stag-hw-parse": {"fixed": True, "on": False},
        "scatter-gather": {"fixed": False, "on": True},
        "tcp-segmentation-offload": {"fixed": False, "on": True},
        "tls-hw-record": {"fixed": True, "on": False},
        "tls-hw-rx-offload": {"fixed": True, "on": False},
        "tls-hw-tx-offload": {"fixed": True, "on": False},
        "tx-checksum-fcoe-crc": {"fixed": True, "on": False},
        "tx-checksum-ip-generic": {"fixed": False, "on": True},
        "tx-checksum-ipv4": {"fixed": True, "on": False},
        "tx-checksum-ipv6": {"fixed": True, "on": False},
        "tx-checksum-sctp": {"fixed": True, "on": False},
        "tx-checksumming": {"fixed": False, "on": True},
        "tx-esp-segmentation": {"fixed": True, "on": False},
        "tx-fcoe-segmentation": {"fixed": True, "on": False},
        "tx-gre-csum-segmentation": {"fixed": True, "on": False},
        "tx-gre-segmentation": {"fixed": True, "on": False},
        "tx-gso-partial": {"fixed": True, "on": False},
        "tx-gso-robust": {"fixed": True, "on": True},
        "tx-ipxip4-segmentation": {"fixed": True, "on": False},
        "tx-ipxip6-segmentation": {"fixed": True, "on": False},
        "tx-lockless": {"fixed": True, "on": False},
        "tx-nocache-copy": {"fixed": False, "on": False},
        "tx-scatter-gather": {"fixed": False, "on": True},
        "tx-scatter-gather-fraglist": {"fixed": True, "on": False},
        "tx-sctp-segmentation": {"fixed": True, "on": False},
        "tx-tcp-ecn-segmentation": {"fixed": False, "on": True},
        "tx-tcp-mangleid-segmentation": {"fixed": False, "on": False},
        "tx-tcp-segmentation": {"fixed": False, "on": True},
        "tx-tcp6-segmentation": {"fixed": False, "on": True},
        "tx-udp-segmentation": {"fixed": True, "on": False},
        "tx-udp_tnl-csum-segmentation": {"fixed": True, "on": False},
        "tx-udp_tnl-segmentation": {"fixed": True, "on": False},
        "tx-vlan-offload": {"fixed": True, "on": False},
        "tx-vlan-stag-hw-insert": {"fixed": True, "on": False},
        "udp-fragmentation-offload": {"fixed": False, "on": False},
        "vlan-challenged": {"fixed": True, "on": False},
    }

    with patch(
        "salt.modules.ethtool._ethtool_command",
        MagicMock(return_value=features_ret.splitlines()),
    ):
        ret = ethtool.show_features("eth0")

    assert expected == ret


def test_ethtool_set_pause():
    with patch("salt.modules.ethtool._ethtool_command", MagicMock(return_value="")):
        with pytest.raises(CommandExecutionError):
            ethtool.set_pause("eth0", not_there=False)
        ret = ethtool.set_pause("eth0", autoneg=False)

    assert ret is True


def test_ethtool_set_feature():
    with patch("salt.modules.ethtool._ethtool_command", MagicMock(return_value="")):
        with pytest.raises(CommandExecutionError):
            ethtool.set_feature("eth0", not_there=False)
        ret = ethtool.set_feature("eth0", sg=False)

    assert ret is True



# Python libs

import pytest

# Salt libs
import salt.beacons.service as service_beacon
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {service_beacon: {"__context__": {}, "__salt__": {}}}


def test_non_list_config():
    config = {}

    ret = service_beacon.validate(config)

    assert ret == (False, "Configuration for service beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = service_beacon.validate(config)

    assert ret == (False, "Configuration for service beacon requires services.")


def test_validate_config_services_none():
    config = [{"services": None}]

    ret = service_beacon.validate(config)

    assert ret == (
        False,
        "Services configuration item for service beacon must be a dictionary.",
    )


def test_validate_config_services_list():
    config = [{"services": [{"sshd": {}}]}]

    ret = service_beacon.validate(config)

    assert ret == (
        False,
        "Services configuration item for service beacon must be a dictionary.",
    )


def test_validate_config_services_valid():
    config = [{"services": {"sshd": {}}}]

    ret = service_beacon.validate(config)

    assert ret == (True, "Valid beacon configuration")


def test_service_running():
    with patch.dict(
        service_beacon.__salt__, {"service.status": MagicMock(return_value=True)}
    ):
        config = [{"services": {"salt-master": {}}}]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        with patch.dict(service_beacon.LAST_STATUS, {}):
            ret = service_beacon.beacon(config)
            assert ret == [
                {
                    "service_name": "salt-master",
                    "tag": "salt-master",
                    "salt-master": {"running": True},
                }
            ]

        # When onchangeonly is True and emitatstartup is False ,
        # we should not see a return when the beacon is run.
        config = [
            {
                "services": {
                    "salt-master": {"emitatstartup": False, "onchangeonly": True}
                }
            }
        ]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        # The return is empty because the beacon did not run
        # but the LAST_STATUS should contain the last status
        # for the service.
        with patch.dict(service_beacon.LAST_STATUS, {}):
            ret = service_beacon.beacon(config)
            assert "salt-master" in service_beacon.LAST_STATUS
            assert ret == []

        # When onchangeonly is True and emitatstartup is
        # the default value True, we should see a return
        # when the beacon is run.
        config = [{"services": {"salt-master": {"onchangeonly": True}}}]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        with patch.dict(service_beacon.LAST_STATUS, {}):
            ret = service_beacon.beacon(config)
            assert ret == [
                {
                    "service_name": "salt-master",
                    "tag": "salt-master",
                    "salt-master": {"running": True},
                }
            ]

        # LAST_STATUS has service name and status has not changed
        config = [{"services": {"salt-master": {"onchangeonly": True}}}]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        mock_ret_dict = {}
        mock_ret_dict["salt-master"] = {"running": True}

        with patch.dict(service_beacon.LAST_STATUS, mock_ret_dict):
            ret = service_beacon.beacon(config)
            assert ret == []

        # LAST_STATUS has service name and status has changed
        config = [{"services": {"salt-master": {"onchangeonly": True}}}]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        mock_ret_dict = {}
        mock_ret_dict["salt-master"] = {"running": False}

        with patch.dict(service_beacon.LAST_STATUS, mock_ret_dict):
            ret = service_beacon.beacon(config)
            assert ret == [
                {
                    "service_name": "salt-master",
                    "tag": "salt-master",
                    "salt-master": {"running": True},
                }
            ]

        # When onchangeonly is True and emitatstartup is True,
        # we should see a return when the beacon is run.
        config = [
            {"services": {"salt-master": {"emitatstartup": True, "onchangeonly": True}}}
        ]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        with patch.dict(service_beacon.LAST_STATUS, {}):
            ret = service_beacon.beacon(config)
            assert ret == [
                {
                    "salt-master": {"running": True},
                    "service_name": "salt-master",
                    "tag": "salt-master",
                }
            ]

            ret = service_beacon.beacon(config)
            assert ret == []


def test_service_not_running():
    with patch.dict(
        service_beacon.__salt__, {"service.status": MagicMock(return_value=False)}
    ):
        config = [{"services": {"salt-master": {}}}]

        ret = service_beacon.validate(config)

        assert ret == (True, "Valid beacon configuration")

        ret = service_beacon.beacon(config)
        assert ret == [
            {
                "service_name": "salt-master",
                "tag": "salt-master",
                "salt-master": {"running": False},
            }
        ]



"""
    :codeauthor: Andrew Colin Kissa <andrew@topdog.za.net>
"""

import pytest

import salt.states.postgres_language as postgres_language
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {postgres_language: {}}


def test_present_existing():
    """
    Test present, language is already present in database
    """
    name = "plpgsql"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_language_list = MagicMock(return_value={"plpgsql": name})
    with patch.dict(
        postgres_language.__salt__,
        {"postgres.language_list": mock_language_list},
    ):
        comt = f"Language {name} is already installed"
        ret.update({"comment": comt, "result": True})
        assert postgres_language.present(name, "testdb") == ret


def test_present_non_existing_pass():
    """
    Test present, language not present in database - pass
    """
    name = "plpgsql"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_true = MagicMock(return_value=True)
    mock_empty_language_list = MagicMock(return_value={})
    with patch.dict(
        postgres_language.__salt__,
        {
            "postgres.language_list": mock_empty_language_list,
            "postgres.language_create": mock_true,
        },
    ):
        with patch.dict(postgres_language.__opts__, {"test": True}):
            comt = f"Language {name} is set to be installed"
            ret.update({"comment": comt, "result": None})
            assert postgres_language.present(name, "testdb") == ret

        with patch.dict(postgres_language.__opts__, {"test": False}):
            comt = f"Language {name} has been installed"
            ret.update(
                {"comment": comt, "result": True, "changes": {"plpgsql": "Present"}}
            )
            assert postgres_language.present(name, "testdb") == ret


def test_present_non_existing_fail():
    """
    Test present, language not present in database - fail
    """
    name = "plpgsql"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_false = MagicMock(return_value=False)
    mock_empty_language_list = MagicMock(return_value={})
    with patch.dict(
        postgres_language.__salt__,
        {
            "postgres.language_list": mock_empty_language_list,
            "postgres.language_create": mock_false,
        },
    ):
        with patch.dict(postgres_language.__opts__, {"test": True}):
            comt = f"Language {name} is set to be installed"
            ret.update({"comment": comt, "result": None})
            assert postgres_language.present(name, "testdb") == ret

        with patch.dict(postgres_language.__opts__, {"test": False}):
            comt = f"Failed to install language {name}"
            ret.update({"comment": comt, "result": False})
            assert postgres_language.present(name, "testdb") == ret


def test_absent_existing():
    """
    Test absent, language present in database
    """
    name = "plpgsql"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_true = MagicMock(return_value=True)
    with patch.dict(
        postgres_language.__salt__,
        {"postgres.language_exists": mock_true, "postgres.language_remove": mock_true},
    ):
        with patch.dict(postgres_language.__opts__, {"test": True}):
            comt = f"Language {name} is set to be removed"
            ret.update({"comment": comt, "result": None})
            assert postgres_language.absent(name, "testdb") == ret

        with patch.dict(postgres_language.__opts__, {"test": False}):
            comt = f"Language {name} has been removed"
            ret.update(
                {"comment": comt, "result": True, "changes": {"plpgsql": "Absent"}}
            )
            assert postgres_language.absent(name, "testdb") == ret


def test_absent_non_existing():
    """
    Test absent, language not present in database
    """
    name = "plpgsql"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_false = MagicMock(return_value=False)
    with patch.dict(
        postgres_language.__salt__, {"postgres.language_exists": mock_false}
    ):
        with patch.dict(postgres_language.__opts__, {"test": True}):
            comt = f"Language {name} is not present so it cannot be removed"
            ret.update({"comment": comt, "result": True})
            assert postgres_language.absent(name, "testdb") == ret
